@@ -1,6 +1,6 @@
 /*
  * Beagle board emulation. http://beagleboard.org/
- * 
+ *
  * Copyright (c) 2009 Nokia Corporation
  *
  * This program is free software; you can redistribute it and/or
@@ -30,6 +30,7 @@
 #include "flash.h"
 #include "sysbus.h"
 #include "blockdev.h"
+#include "exec-memory.h"
 
 #define BEAGLE_NAND_CS       0
 #define BEAGLE_SMC_CS        1
@@ -58,56 +59,42 @@ static void beagle_common_init(ram_addr_t ram_size,
                         const char *initrd_filename,
                         int cpu_model)
 {
-    struct beagle_s *s = (struct beagle_s *) qemu_mallocz(sizeof(*s));
+    MemoryRegion *sysmem = get_system_memory();
+    struct beagle_s *s = (struct beagle_s *) g_malloc0(sizeof(*s));
     DriveInfo *dmtd = drive_get(IF_MTD, 0, 0);
     DriveInfo *dsd  = drive_get(IF_SD, 0, 0);
-    
+
     if (!dmtd && !dsd) {
         hw_error("%s: SD or NAND image required", __FUNCTION__);
     }
 #if MAX_SERIAL_PORTS < 1
 #error MAX_SERIAL_PORTS must be at least 1!
 #endif
-    s->cpu = omap3_mpu_init(cpu_model, 1, ram_size,
+    s->cpu = omap3_mpu_init(sysmem, cpu_model, ram_size,
                             NULL, NULL, serial_hds[0], NULL);
 
     s->nand = nand_init(dmtd ? dmtd->bdrv : NULL, NAND_MFR_MICRON, 0xba);
     nand_setpins(s->nand, 0, 0, 0, 1, 0); /* no write-protect */
-    omap_gpmc_attach(s->cpu->gpmc, BEAGLE_NAND_CS, s->nand, 0, 2);
+    omap_gpmc_attach_nand(s->cpu->gpmc, BEAGLE_NAND_CS, s->nand);
 
     if (dsd) {
         omap3_mmc_attach(s->cpu->omap3_mmc[0], dsd->bdrv, 0, 0);
     }
 
-    s->twl4030 = twl4030_init(omap_i2c_bus(s->cpu->i2c, 0),
-                              s->cpu->irq[0][OMAP_INT_3XXX_SYS_NIRQ],
+    s->twl4030 = twl4030_init(omap_i2c_bus(s->cpu->i2c[0]),
+                              qdev_get_gpio_in(s->cpu->ih[0],
+                                               OMAP_INT_3XXX_SYS_NIRQ),
                               NULL, NULL);
-    int i;
-    for (i = 0; i < nb_nics; i++) {
-        if (!nd_table[i].model || !strcmp(nd_table[i].model, "smc91c111")) {
-            break;
-        }
-    }
     if (cpu_model == omap3430) {
         qemu_set_irq(qdev_get_gpio_in(s->cpu->gpio, BEAGLE_GPIO_ID1),1);
         qemu_set_irq(qdev_get_gpio_in(s->cpu->gpio, BEAGLE_GPIO_ID3),1);
     }
-    if (i < nb_nics) {
-        s->smc = qdev_create(NULL, "smc91c111");
-        qdev_set_nic_properties(s->smc, &nd_table[i]);
-        qdev_init_nofail(s->smc);
-        sysbus_connect_irq(sysbus_from_qdev(s->smc), 0,
-                           qdev_get_gpio_in(s->cpu->gpio, 54));
-    } else {
-        hw_error("%s: no NIC for smc91c111\n", __FUNCTION__);
-    }
-    omap_gpmc_attach(s->cpu->gpmc, BEAGLE_SMC_CS, s->smc, 0, 0);
 
     /* Wire up an I2C slave which returns EDID monitor information;
      * newer Linux kernels won't turn on the display unless they
      * detect a monitor over DDC.
      */
-    s->ddc = i2c_create_slave(omap_i2c_bus(s->cpu->i2c, 2), "i2c-ddc", 0x50);
+    s->ddc = i2c_create_slave(omap_i2c_bus(s->cpu->i2c[2]), "i2c-ddc", 0x50);
 
     omap_lcd_panel_attach(s->cpu->dss);
 }
