@@ -35,6 +35,7 @@
 #include "loader.h"
 #include "blockdev.h"
 #include "sysbus.h"
+#include "exec-memory.h"
 
 //#define MIPID_DEBUG
 
@@ -174,12 +175,21 @@ static void n8x0_nand_setup(struct n800_s *s)
     char *otp_region;
     DriveInfo *dinfo;
 
-    dinfo = drive_get(IF_MTD, 0, 0);
+    s->nand = qdev_create(NULL, "onenand");
+    qdev_prop_set_uint16(s->nand, "manufacturer_id", NAND_MFR_SAMSUNG);
     /* Either 0x40 or 0x48 are OK for the device ID */
-    s->nand = onenand_init(dinfo ? dinfo->bdrv : 0,
-                           NAND_MFR_SAMSUNG, 0x48, 0, 1,
-                           qdev_get_gpio_in(s->cpu->gpio, N8X0_ONENAND_GPIO));
-    omap_gpmc_attach(s->cpu->gpmc, N8X0_ONENAND_CS, s->nand, 0, OMAP_GPMC_NOR);
+    qdev_prop_set_uint16(s->nand, "device_id", 0x48);
+    qdev_prop_set_uint16(s->nand, "version_id", 0);
+    qdev_prop_set_int32(s->nand, "shift", 1);
+    dinfo = drive_get(IF_MTD, 0, 0);
+    if (dinfo && dinfo->bdrv) {
+        qdev_prop_set_drive_nofail(s->nand, "drive", dinfo->bdrv);
+    }
+    qdev_init_nofail(s->nand);
+    sysbus_connect_irq(sysbus_from_qdev(s->nand), 0,
+                       qdev_get_gpio_in(s->cpu->gpio, N8X0_ONENAND_GPIO));
+    omap_gpmc_attach(s->cpu->gpmc, N8X0_ONENAND_CS,
+                     sysbus_mmio_get_region(sysbus_from_qdev(s->nand), 0));
     otp_region = onenand_raw_otp(s->nand);
 
     memcpy(otp_region + 0x000, n8x0_cal_wlan_mac, sizeof(n8x0_cal_wlan_mac));
@@ -191,11 +201,15 @@ static void n8x0_i2c_setup(struct n800_s *s)
 {
     DeviceState *dev;
     qemu_irq tmp_irq = qdev_get_gpio_in(s->cpu->gpio, N8X0_TMP105_GPIO);
-    i2c_bus *i2c = omap_i2c_bus(s->cpu->i2c, 0);
+    i2c_bus *i2c = omap_i2c_bus(s->cpu->i2c[0]);
 
     /* Attach a menelaus PM chip */
     dev = i2c_create_slave(i2c, "twl92230", N8X0_MENELAUS_ADDR);
-    qdev_connect_gpio_out(dev, 3, s->cpu->irq[0][OMAP_INT_24XX_SYS_NIRQ]);
+    qdev_connect_gpio_out(dev, 3,
+                          qdev_get_gpio_in(s->cpu->ih[0],
+                                           OMAP_INT_24XX_SYS_NIRQ));
+
+    qemu_system_powerdown = qdev_get_gpio_in(dev, 3);
 
     /* Attach a TMP105 PM chip (A0 wired to ground) */
     dev = i2c_create_slave(i2c, "tmp105", N8X0_TMP105_ADDR);
@@ -379,7 +393,7 @@ static void n810_kbd_setup(struct n800_s *s)
 
     /* Attach the LM8322 keyboard to the I2C bus,
      * should happen in n8x0_i2c_setup and s->kbd be initialised here.  */
-    s->kbd = i2c_create_slave(omap_i2c_bus(s->cpu->i2c, 0),
+    s->kbd = i2c_create_slave(omap_i2c_bus(s->cpu->i2c[0]),
                            "lm8323", N810_LM8323_ADDR);
     qdev_connect_gpio_out(s->kbd, 0, kbd_irq);
 }
@@ -415,9 +429,9 @@ static void mipid_reset(DeviceState *qdev)
 {
     struct mipid_s *s = FROM_SPI_DEVICE(struct mipid_s,
                                         SPI_DEVICE_FROM_QDEV(qdev));
-    
-    //if (!s->sleep)
-    //    fprintf(stderr, "%s: Display off\n", __FUNCTION__);
+    if (!s->sleep) {
+        TRACE_MIPID("display off");
+    }
 
     s->pm = 0;
     s->cmd = 0;
@@ -786,17 +800,28 @@ static int mipid_init(SPIDevice *spidev)
     return 0;
 }
 
-static SPIDeviceInfo mipid_info = {
-    .init = mipid_init,
-    .txrx = mipid_txrx,
-    .qdev.name = "lcd_mipid",
-    .qdev.size = sizeof(struct mipid_s),
-    .qdev.reset = mipid_reset,
-    .qdev.props = (Property[]) {
-        DEFINE_PROP_UINT32("id", struct mipid_s, id, 0),
-        DEFINE_PROP_UINT8("n900", struct mipid_s, n900, 0),
-        DEFINE_PROP_END_OF_LIST()
-    }
+static Property mipid_properties[] = {
+    DEFINE_PROP_UINT32("id", struct mipid_s, id, 0),
+    DEFINE_PROP_UINT8("n900", struct mipid_s, n900, 0),
+    DEFINE_PROP_END_OF_LIST()
+};
+
+static void mipid_class_init(ObjectClass *klass, void *data)
+{
+    DeviceClass *dc = DEVICE_CLASS(klass);
+    SPIDeviceClass *k = SPI_DEVICE_CLASS(klass);
+
+    k->init = mipid_init;
+    k->txrx = mipid_txrx;
+    dc->reset = mipid_reset;
+    dc->props = mipid_properties;
+}
+
+static TypeInfo mipid_info = {
+    .name = "lcd_mipid",
+    .parent = TYPE_SPI_DEVICE,
+    .instance_size = sizeof(struct mipid_s),
+    .class_init = mipid_class_init,
 };
 
 static void n8x0_spi_setup(struct n800_s *s)
@@ -846,10 +871,10 @@ static void n800_dss_init(struct rfbi_chip_s *chip)
     chip->write(chip->opaque, 1, 0x01);		/* Input Data Format */
     chip->write(chip->opaque, 1, 0x01);		/* Data Source Select */
 
-    fb_blank = memset(qemu_malloc(800 * 480 * 2), 0xff, 800 * 480 * 2);
+    fb_blank = memset(g_malloc(800 * 480 * 2), 0xff, 800 * 480 * 2);
     /* Display Memory Data Port */
     chip->block(chip->opaque, 1, fb_blank, 800 * 480 * 2, 800);
-    qemu_free(fb_blank);
+    g_free(fb_blank);
 }
 
 static void n8x0_dss_setup(struct n800_s *s)
@@ -893,14 +918,19 @@ static void n8x0_uart_setup(struct n800_s *s)
 
 static void n8x0_usb_setup(struct n800_s *s)
 {
+    SysBusDevice *dev;
     s->usb = qdev_create(NULL, "tusb6010");
-    sysbus_connect_irq(sysbus_from_qdev(s->usb), 0,
+    dev = sysbus_from_qdev(s->usb);
+    qdev_init_nofail(s->usb);
+    sysbus_connect_irq(dev, 0,
                        qdev_get_gpio_in(s->cpu->gpio, N8X0_TUSB_INT_GPIO));
     /* Using the NOR interface */
-    omap_gpmc_attach(s->cpu->gpmc, N8X0_USB_SYNC_CS, s->usb, 0, OMAP_GPMC_NOR);
-    omap_gpmc_attach(s->cpu->gpmc, N8X0_USB_ASYNC_CS, s->usb, 1, OMAP_GPMC_NOR);
+    omap_gpmc_attach(s->cpu->gpmc, N8X0_USB_ASYNC_CS,
+                     sysbus_mmio_get_region(dev, 0));
+    omap_gpmc_attach(s->cpu->gpmc, N8X0_USB_SYNC_CS,
+                     sysbus_mmio_get_region(dev, 1));
     qdev_connect_gpio_out(s->cpu->gpio, N8X0_TUSB_ENABLE_GPIO,
-                          qdev_get_gpio_in(s->usb, 0)); /* tusb_power */
+                          qdev_get_gpio_in(s->usb, 0)); /* tusb_pwr */
 }
 
 /* Setup done before the main bootloader starts by some early setup code
@@ -1395,11 +1425,12 @@ static void n8x0_init(ram_addr_t ram_size, const char *boot_device,
                 const char *kernel_cmdline, const char *initrd_filename,
                 const char *cpu_model, struct arm_boot_info *binfo, int model)
 {
-    struct n800_s *s = (struct n800_s *) qemu_mallocz(sizeof(*s));
+    MemoryRegion *sysmem = get_system_memory();
+    struct n800_s *s = (struct n800_s *) g_malloc0(sizeof(*s));
     int sdram_size = binfo->ram_size;
     DisplayState *ds;
 
-    s->cpu = omap2420_mpu_init(sdram_size, cpu_model);
+    s->cpu = omap2420_mpu_init(sysmem, sdram_size, cpu_model);
 
     /* Setup peripherals
      *
@@ -1577,7 +1608,7 @@ static QEMUMachine n810_machine = {
 #define TRACE_LIS302DL(...)
 #endif
 
-static uint32_t ssi_read(void *opaque, target_phys_addr_t addr)
+static uint64_t ssi_read(void *opaque, target_phys_addr_t addr, unsigned size)
 {
     switch (addr) {
         case 0x00: /* REVISION */
@@ -1591,25 +1622,20 @@ static uint32_t ssi_read(void *opaque, target_phys_addr_t addr)
     return 0;
 }
 
-static void ssi_write(void *opaque, target_phys_addr_t addr, uint32_t value)
+static void ssi_write(void *opaque, target_phys_addr_t addr, uint64_t value,
+                      unsigned size)
 {
     //printf("%s: addr=" OMAP_FMT_plx ", value=0x%08x\n", __FUNCTION__, addr, value);
 }
 
-static CPUReadMemoryFunc *ssi_read_func[] = {
-    ssi_read,
-    ssi_read,
-    ssi_read,
-};
-
-static CPUWriteMemoryFunc *ssi_write_func[] = {
-    ssi_write,
-    ssi_write,
-    ssi_write,
+static const MemoryRegionOps ssi_ops = {
+    .read = ssi_read,
+    .write = ssi_write,
+    .endianness = DEVICE_NATIVE_ENDIAN,
 };
 
 typedef struct LIS302DLState_s {
-    i2c_slave i2c;
+    I2CSlave i2c;
     int firstbyte;
     uint8_t reg;
 
@@ -1749,28 +1775,6 @@ static void lis302dl_step(void *opaque, int axis, int high, int activate)
     }
 }
 
-static int lis302dl_change(DeviceState *dev, const char *target,
-                           const char *arg)
-{
-    LIS302DLState *s = (LIS302DLState *)dev;
-    int axis;
-    if (!strcmp(target, "x")) {
-        axis = 0;
-    } else if (!strcmp(target, "y")) {
-        axis = 1;
-    } else if (!strcmp(target, "z")) {
-        axis = 2;
-    } else {
-        return -1;
-    }
-    int value = 0;
-    if (sscanf(arg, "%d", &value) != 1) {
-        return -1;
-    }
-    lis302dl_trigger(s, axis, value);
-    return 0;
-}
-
 static void lis302dl_reset(DeviceState *ds)
 {
     LIS302DLState *s = FROM_I2C_SLAVE(LIS302DLState, I2C_SLAVE_FROM_QDEV(ds));
@@ -1796,7 +1800,7 @@ static void lis302dl_reset(DeviceState *ds)
     lis302dl_interrupt_update(s);
 }
 
-static void lis302dl_event(i2c_slave *i2c, enum i2c_event event)
+static void lis302dl_event(I2CSlave *i2c, enum i2c_event event)
 {
     LIS302DLState *s = FROM_I2C_SLAVE(LIS302DLState, i2c);
     if (event == I2C_START_SEND)
@@ -1853,7 +1857,7 @@ static uint8_t lis302dl_readcoord(LIS302DLState *s, int coord)
     return (uint8_t)v;
 }
 
-static int lis302dl_rx(i2c_slave *i2c)
+static int lis302dl_rx(I2CSlave *i2c)
 {
     LIS302DLState *s = FROM_I2C_SLAVE(LIS302DLState, i2c);
     int value = -1;
@@ -1967,7 +1971,7 @@ static int lis302dl_rx(i2c_slave *i2c)
     return value;
 }
 
-static int lis302dl_tx(i2c_slave *i2c, uint8_t data)
+static int lis302dl_tx(I2CSlave *i2c, uint8_t data)
 {
     LIS302DLState *s = FROM_I2C_SLAVE(LIS302DLState, i2c);
     if (s->firstbyte) {
@@ -2044,7 +2048,7 @@ static int lis302dl_tx(i2c_slave *i2c, uint8_t data)
     return 1;
 }
 
-static int lis302dl_init(i2c_slave *i2c)
+static int lis302dl_init(I2CSlave *i2c)
 {
     LIS302DLState *s = FROM_I2C_SLAVE(LIS302DLState, i2c);
     s->axis_max = 58;
@@ -2053,25 +2057,39 @@ static int lis302dl_init(i2c_slave *i2c)
     return 0;
 }
 
-static I2CSlaveInfo lis302dl_info = {
-    .qdev.name = "lis302dl",
-    .qdev.size = sizeof(LIS302DLState),
-    .qdev.reset = lis302dl_reset,
-    .qdev.change = lis302dl_change,
-    .qdev.props = (Property[]) {
-        DEFINE_PROP_INT32("x", LIS302DLState, x, 0),
-        DEFINE_PROP_INT32("y", LIS302DLState, y, 0),
-        DEFINE_PROP_INT32("z", LIS302DLState, z, 0),
-        DEFINE_PROP_END_OF_LIST(),
-    },
-    .init = lis302dl_init,
-    .event = lis302dl_event,
-    .recv = lis302dl_rx,
-    .send = lis302dl_tx
+/* TODO: ideally x, y, z should be runtime modifiable properties,
+ * which can be set by calling
+ *     lis302dl_trigger(s, axis, value);
+ * where axis is 0,1,2 for x,y,z
+ */
+static Property lis302dl_properties[] = {
+    DEFINE_PROP_INT32("x", LIS302DLState, x, 0),
+    DEFINE_PROP_INT32("y", LIS302DLState, y, 0),
+    DEFINE_PROP_INT32("z", LIS302DLState, z, 0),
+    DEFINE_PROP_END_OF_LIST(),
+};
+
+static void lis302dl_class_init(ObjectClass *klass, void *data)
+{
+    DeviceClass *dc = DEVICE_CLASS(klass);
+    I2CSlaveClass *k = I2C_SLAVE_CLASS(klass);
+    k->init = lis302dl_init;
+    k->event = lis302dl_event;
+    k->recv = lis302dl_rx;
+    k->send = lis302dl_tx;
+    dc->props = lis302dl_properties;
+    dc->reset = lis302dl_reset;
+}
+
+static TypeInfo lis302dl_info = {
+    .name = "lis302dl",
+    .parent = TYPE_I2C_SLAVE,
+    .instance_size = sizeof(LIS302DLState),
+    .class_init = lis302dl_class_init,
 };
 
 typedef struct BQ2415XState_s {
-    i2c_slave i2c;
+    I2CSlave i2c;
     int firstbyte;
     uint8 reg;
     
@@ -2095,14 +2113,14 @@ static void bq2415x_reset(DeviceState *ds)
     s->tcc = 0xa1; // 89
 }
 
-static void bq2415x_event(i2c_slave *i2c, enum i2c_event event)
+static void bq2415x_event(I2CSlave *i2c, enum i2c_event event)
 {
     BQ2415XState *s = FROM_I2C_SLAVE(BQ2415XState, i2c);
     if (event == I2C_START_SEND)
         s->firstbyte = 1;
 }
 
-static int bq2415x_rx(i2c_slave *i2c)
+static int bq2415x_rx(I2CSlave *i2c)
 {
     BQ2415XState *s = FROM_I2C_SLAVE(BQ2415XState, i2c);
     int value = -1;
@@ -2137,7 +2155,7 @@ static int bq2415x_rx(i2c_slave *i2c)
     return value;
 }
 
-static int bq2415x_tx(i2c_slave *i2c, uint8_t data)
+static int bq2415x_tx(I2CSlave *i2c, uint8_t data)
 {
     BQ2415XState *s = FROM_I2C_SLAVE(BQ2415XState, i2c);
     if (s->firstbyte) {
@@ -2171,27 +2189,37 @@ static int bq2415x_tx(i2c_slave *i2c, uint8_t data)
     return 1;
 }
 
-static int bq2415x_init(i2c_slave *i2c)
+static int bq2415x_init(I2CSlave *i2c)
 {
     return 0;
 }
 
-static I2CSlaveInfo bq2415x_info = {
-    .qdev.name = "bq2415x",
-    .qdev.size = sizeof(BQ2415XState),
-    .qdev.reset = bq2415x_reset,
-    .qdev.props = (Property[]) {
-        DEFINE_PROP_UINT8("id", BQ2415XState, id, 0x49),
-        DEFINE_PROP_END_OF_LIST(),
-    },
-    .init = bq2415x_init,
-    .event = bq2415x_event,
-    .recv = bq2415x_rx,
-    .send = bq2415x_tx
+static Property bq2415x_properties[] = {
+    DEFINE_PROP_UINT8("id", BQ2415XState, id, 0x49),
+    DEFINE_PROP_END_OF_LIST(),
+};
+
+static void bq2415x_class_init(ObjectClass *klass, void *data)
+{
+    DeviceClass *dc = DEVICE_CLASS(klass);
+    I2CSlaveClass *k = I2C_SLAVE_CLASS(klass);
+    k->init = bq2415x_init;
+    k->event = bq2415x_event;
+    k->recv = bq2415x_rx;
+    k->send = bq2415x_tx;
+    dc->props = bq2415x_properties;
+    dc->reset = bq2415x_reset;
+}
+
+static TypeInfo bq2415x_info = {
+    .name = "bq2415x",
+    .parent = TYPE_I2C_SLAVE,
+    .instance_size = sizeof(BQ2415XState),
+    .class_init = bq2415x_class_init,
 };
 
 typedef struct tpa6130_s {
-    i2c_slave i2c;
+    I2CSlave i2c;
     int firstbyte;
     int reg;
     uint8_t data[3];
@@ -2205,14 +2233,14 @@ static void tpa6130_reset(DeviceState *ds)
     memset(s->data, 0, sizeof(s->data));
 }
 
-static void tpa6130_event(i2c_slave *i2c, enum i2c_event event)
+static void tpa6130_event(I2CSlave *i2c, enum i2c_event event)
 {
     TPA6130State *s = FROM_I2C_SLAVE(TPA6130State, i2c);
     if (event == I2C_START_SEND)
         s->firstbyte = 1;
 }
 
-static int tpa6130_rx(i2c_slave *i2c)
+static int tpa6130_rx(I2CSlave *i2c)
 {
     TPA6130State *s = FROM_I2C_SLAVE(TPA6130State, i2c);
     int value = 0;
@@ -2233,7 +2261,7 @@ static int tpa6130_rx(i2c_slave *i2c)
     return value;
 }
 
-static int tpa6130_tx(i2c_slave *i2c, uint8_t data)
+static int tpa6130_tx(I2CSlave *i2c, uint8_t data)
 {
     TPA6130State *s = FROM_I2C_SLAVE(TPA6130State, i2c);
     if (s->firstbyte) {
@@ -2264,20 +2292,28 @@ static void tpa6130_irq(void *opaque, int n, int level)
     }
 }
 
-static int tpa6130_init(i2c_slave *i2c)
+static int tpa6130_init(I2CSlave *i2c)
 {
     qdev_init_gpio_in(&i2c->qdev, tpa6130_irq, 1);
     return 0;
 }
 
-static I2CSlaveInfo tpa6130_info = {
-    .qdev.name = "tpa6130",
-    .qdev.size = sizeof(TPA6130State),
-    .qdev.reset = tpa6130_reset,
-    .init = tpa6130_init,
-    .event = tpa6130_event,
-    .recv = tpa6130_rx,
-    .send = tpa6130_tx
+static void tpa6130_class_init(ObjectClass *klass, void *data)
+{
+    DeviceClass *dc = DEVICE_CLASS(klass);
+    I2CSlaveClass *k = I2C_SLAVE_CLASS(klass);
+    k->init = tpa6130_init;
+    k->event = tpa6130_event;
+    k->recv = tpa6130_rx;
+    k->send = tpa6130_tx;
+    dc->reset = tpa6130_reset;
+}
+
+static TypeInfo tpa6130_info = {
+    .name = "tpa6130",
+    .parent = TYPE_I2C_SLAVE,
+    .instance_size = sizeof(TPA6130State),
+    .class_init = tpa6130_class_init,
 };
 
 struct n900_s {
@@ -2473,7 +2509,9 @@ static void n900_init(ram_addr_t ram_size,
                       const char *initrd_filename,
                       const char *cpu_model)
 {
-    struct n900_s *s = qemu_mallocz(sizeof(*s));
+    MemoryRegion *sysmem = get_system_memory();
+    MemoryRegion *ssi_iomem = g_new(MemoryRegion, 1);
+    struct n900_s *s = g_malloc0(sizeof(*s));
     DriveInfo *dmtd = drive_get(IF_MTD, 0, 0);
     DriveInfo *dsd  = drive_get(IF_SD, 0, 0);
 
@@ -2483,7 +2521,7 @@ static void n900_init(ram_addr_t ram_size,
 #if MAX_SERIAL_PORTS < 3
 #error MAX_SERIAL_PORTS must be at least 3!
 #endif
-    s->cpu = omap3_mpu_init(omap3430, 1, N900_SDRAM_SIZE,
+    s->cpu = omap3_mpu_init(sysmem, omap3430, N900_SDRAM_SIZE,
                             serial_hds[1], serial_hds[2],
                             serial_hds[0], NULL);
     omap_lcd_panel_attach(s->cpu->dss);
@@ -2502,10 +2540,20 @@ static void n900_init(ram_addr_t ram_size,
     qdev_prop_set_uint8(s->mipid, "n900", 1);
     qdev_init_nofail(s->mipid);
 
-    s->nand = onenand_init(dmtd ? dmtd->bdrv : NULL,
-                           NAND_MFR_SAMSUNG, 0x40, 0x121, 1,
-                           qdev_get_gpio_in(s->cpu->gpio, N900_ONENAND_GPIO));
-    omap_gpmc_attach(s->cpu->gpmc, 0, s->nand, 0, 0);
+
+    s->nand = qdev_create(NULL, "onenand");
+    qdev_prop_set_uint16(s->nand, "manufacturer_id", NAND_MFR_SAMSUNG);
+    qdev_prop_set_uint16(s->nand, "device_id", 0x40);
+    qdev_prop_set_uint16(s->nand, "version_id", 0x121);
+    qdev_prop_set_int32(s->nand, "shift", 1);
+    if (dmtd && dmtd->bdrv) {
+        qdev_prop_set_drive_nofail(s->nand, "drive", dmtd->bdrv);
+    }
+    qdev_init_nofail(s->nand);
+    sysbus_connect_irq(sysbus_from_qdev(s->nand), 0,
+                       qdev_get_gpio_in(s->cpu->gpio, N900_ONENAND_GPIO));
+    omap_gpmc_attach(s->cpu->gpmc, 0,
+                     sysbus_mmio_get_region(sysbus_from_qdev(s->nand), 0));
 
     if (dsd) {
         omap3_mmc_attach(s->cpu->omap3_mmc[1], dsd->bdrv, 0, 1);
@@ -2515,22 +2563,20 @@ static void n900_init(ram_addr_t ram_size,
         //qemu_irq_raise(omap2_gpio_in_get(s->cpu->gpif, N900_SDCOVER_GPIO));
     }
 
-    cpu_register_physical_memory(0x48058000, 0x3c00,
-                                 cpu_register_io_memory(ssi_read_func,
-                                                        ssi_write_func,
-                                                        0,
-                                                        DEVICE_NATIVE_ENDIAN));
+    memory_region_init_io(ssi_iomem, &ssi_ops, 0, "n900_ssi", 0x3c00);
+    memory_region_add_subregion(sysmem, 0x48058000, ssi_iomem);
 
-    s->twl4030 = twl4030_init(omap_i2c_bus(s->cpu->i2c, 0),
-                              s->cpu->irq[0][OMAP_INT_3XXX_SYS_NIRQ],
+    s->twl4030 = twl4030_init(omap_i2c_bus(s->cpu->i2c[0]),
+                              qdev_get_gpio_in(s->cpu->ih[0],
+                                               OMAP_INT_3XXX_SYS_NIRQ),
                               NULL, n900_twl4030_keymap);
     twl4030_madc_attach(s->twl4030, n900_twl4030_madc_callback);
-    i2c_bus *i2c2 = omap_i2c_bus(s->cpu->i2c, 1);
+    i2c_bus *i2c2 = omap_i2c_bus(s->cpu->i2c[1]);
     s->bq2415x = i2c_create_slave(i2c2, "bq2415x", 0x6b);
     s->tpa6130 = i2c_create_slave(i2c2, "tpa6130", 0x60);
     qdev_connect_gpio_out(s->cpu->gpio, N900_HEADPHONE_EN_GPIO,
                           qdev_get_gpio_in(s->tpa6130, 0));
-    i2c_bus *i2c3 = omap_i2c_bus(s->cpu->i2c, 2);
+    i2c_bus *i2c3 = omap_i2c_bus(s->cpu->i2c[2]);
     s->lis302dl = i2c_create_slave(i2c3, "lis302dl", 0x1d);
     qdev_connect_gpio_out(s->lis302dl, 0,
                           qdev_get_gpio_in(s->cpu->gpio,
@@ -2551,7 +2597,8 @@ static void n900_init(ram_addr_t ram_size,
         qdev_init_nofail(s->smc);
         sysbus_connect_irq(sysbus_from_qdev(s->smc), 0,
                            qdev_get_gpio_in(s->cpu->gpio, 54));
-        omap_gpmc_attach(s->cpu->gpmc, 1, s->smc, 0, 0);
+        omap_gpmc_attach(s->cpu->gpmc, 1,
+                         sysbus_mmio_get_region(sysbus_from_qdev(s->smc), 0));
     } else {
         hw_error("%s: no NIC for smc91c111\n", __FUNCTION__);
     }
@@ -2567,12 +2614,12 @@ static QEMUMachine n900_machine = {
     .init = n900_init,
 };
 
-static void nseries_register_devices(void)
+static void nseries_register_types(void)
 {
-    i2c_register_slave(&bq2415x_info);
-    i2c_register_slave(&tpa6130_info);
-    i2c_register_slave(&lis302dl_info);
-    spi_register_device(&mipid_info);
+    type_register_static(&bq2415x_info);
+    type_register_static(&tpa6130_info);
+    type_register_static(&lis302dl_info);
+    type_register_static(&mipid_info);
 }
 
 static void nseries_machine_init(void)
@@ -2582,5 +2629,5 @@ static void nseries_machine_init(void)
     qemu_register_machine(&n900_machine);
 }
 
-device_init(nseries_register_devices);
+type_init(nseries_register_types);
 machine_init(nseries_machine_init);
