@@ -22,15 +22,6 @@
 #include "omap.h"
 #include "irq.h"
 #include "soc_dma.h"
-#include "hw.h"
-
-//#define OMAP_DMA_DEBUG
-
-#ifdef OMAP_DMA_DEBUG
-#define TRACE(fmt,...) printf("%s:" fmt "\n", __FUNCTION__, ##__VA_ARGS__)
-#else
-#define TRACE(...)
-#endif
 
 struct omap_dma_channel_s {
     /* transfer data */
@@ -107,7 +98,7 @@ struct omap_dma_channel_s {
     int type;
     int suspend;
     int buf_disable;
-    
+
     struct omap_dma_desc_linked_list_s {
         uint32_t ptr;
         uint32_t next;
@@ -117,6 +108,7 @@ struct omap_dma_channel_s {
 
 struct omap_dma_s {
     struct soc_dma_s *dma;
+    MemoryRegion iomem;
 
     struct omap_mpu_state_s *mpu;
     omap_clk clk;
@@ -265,8 +257,9 @@ static void omap_dma_deactivate_channel(struct omap_dma_s *s,
 
     /* Don't deactive the channel if it is synchronized and the DMA request is
        active */
-    if (ch->sync && ch->enable && s->dma->drqst[ch->sync])
+    if (ch->sync && ch->enable && s->dma->drqst[ch->sync]) {
         return;
+    }
 
     if (ch->active) {
         ch->active = 0;
@@ -290,9 +283,8 @@ static int omap_dma_sgl_next(struct omap_dma_s *s,
     int src_valid = (ch->sgl.ptr >> 2) & 3; /* SRC_VALID */
     int dst_valid = ch->sgl.ptr & 3;        /* DEST_VALID */
     if (type < 1 || type > 3) {
-        hw_error("%s: unknown descriptor type %d\n", __FUNCTION__, type);
+        hw_error("%s: unknown descriptor type %d\n", __func__, type);
     }
-    TRACE("loading next descriptor at 0x%08x, type %d", ch->sgl.next, type);
     target_phys_addr_t addr = (target_phys_addr_t)ch->sgl.next;
     ch->sgl.ptr &= ~0xff;
     uint8_t data[4];
@@ -312,44 +304,42 @@ static int omap_dma_sgl_next(struct omap_dma_s *s,
     }
     addr += 8;
     switch (src_valid) {
-        case 0:
-            ch->addr[0] = ch->active_set.src;
-            break;
-        case 1:
-            cpu_physical_memory_read(addr, data, 4);
-            ch->addr[0] = ldl_p(data);
-            addr += 4;
-            break;
-        case 2:
-            break;
-        default:
-            hw_error("%s: unknown src addressing mode %d\n",
-                     __FUNCTION__, src_valid);
-            break;
+    case 0:
+        ch->addr[0] = ch->active_set.src;
+        break;
+    case 1:
+        cpu_physical_memory_read(addr, data, 4);
+        ch->addr[0] = ldl_p(data);
+        addr += 4;
+        break;
+    case 2:
+        break;
+    default:
+        hw_error("%s: unknown src addressing mode %d\n",
+                 __func__, src_valid);
+        break;
     }
     switch (dst_valid) {
-        case 0:
-            ch->addr[1] = ch->active_set.dest;
-            break;
-        case 1:
-            cpu_physical_memory_read(addr, data, 4);
-            ch->addr[1] = ldl_p(data);
-            addr += 4;
-            break;
-        case 2:
-            break;
-        default:
-            hw_error("%s: unknown dest addressing mode %d\n",
-                     __FUNCTION__, dst_valid);
-            break;
+    case 0:
+        ch->addr[1] = ch->active_set.dest;
+        break;
+    case 1:
+        cpu_physical_memory_read(addr, data, 4);
+        ch->addr[1] = ldl_p(data);
+        addr += 4;
+        break;
+    case 2:
+        break;
+    default:
+        hw_error("%s: unknown dest addressing mode %d\n",
+                 __func__, dst_valid);
+        break;
     }
     if (type == 1 || type == 2) {
         cpu_physical_memory_read(addr, data, 4);
         word = ldl_p(data);
         ch->interrupts = word >> 16;
-        TRACE("CICR = 0x%08x", ch->interrupts);
         ch->frames = word & 0xffff;
-        TRACE("CFN = %d", ch->frames);
         cpu_physical_memory_read(addr + 4, data, 4);
         word = ldl_p(data);
         ch->element_index[0] = (int16_t)(word >> 16);
@@ -382,8 +372,9 @@ static void omap_dma_enable_channel(struct omap_dma_s *s,
         /* TODO: theoretically if ch->sync && ch->prefetch &&
          * !s->dma->drqst[ch->sync], we should also activate and fetch
          * from source and then stall until signalled.  */
-        if ((!ch->sync) || s->dma->drqst[ch->sync])
+        if ((!ch->sync) || s->dma->drqst[ch->sync]) {
             omap_dma_activate_channel(s, ch);
+        }
     }
 }
 
@@ -483,26 +474,25 @@ static void omap_dma_end_of_block(struct omap_dma_s *s,
 {
     if (ch->omap_3_1_compatible_disable) {
         omap_dma_disable_channel(s, ch);
-        if (ch->link_enabled)
+        if (ch->link_enabled) {
             omap_dma_enable_channel(s, &s->ch[ch->link_next_ch]);
+        }
     } else {
         int sgl_pause = (ch->sgl.ptr >> 7) & 1; /* PAUSE_LINK_LIST */
-        TRACE("end of block, transfermode=%d, pause=%d",
-              (ch->sgl.ptr >> 8) & 3, sgl_pause);
         if (((ch->sgl.ptr >> 8) & 3) != 1 || /* TRANSFER_MODE != linked list */
             !omap_dma_sgl_next(s, ch) ||     /* end of list */
             sgl_pause) {
-            if (!ch->auto_init)
+            if (!ch->auto_init) {
                 omap_dma_disable_channel(s, ch);
-            else if (ch->repeat || ch->end_prog)
+            } else if (ch->repeat || ch->end_prog) {
                 omap_dma_channel_load(ch);
-            else {
+            } else {
                 ch->waiting_end_prog = 1;
                 omap_dma_deactivate_channel(s, ch);
             }
         }
     }
-    
+
     if (ch->interrupts & END_BLOCK_INTR)
         ch->status |= END_BLOCK_INTR;
 }
@@ -517,7 +507,6 @@ static void omap_dma_transfer_generic(struct soc_dma_ch_s *dma)
     uint16_t status = ch->status;
 #endif
 
-    TRACE("frame %d", a->frame);
     do {
         /* Transfer a single element */
         /* FIXME: check the endianness */
@@ -1570,12 +1559,17 @@ static int omap_dma_sys_read(struct omap_dma_s *s, int offset,
     return 0;
 }
 
-static uint32_t omap_dma_read(void *opaque, target_phys_addr_t addr)
+static uint64_t omap_dma_read(void *opaque, target_phys_addr_t addr,
+                              unsigned size)
 {
     struct omap_dma_s *s = (struct omap_dma_s *) opaque;
     int reg, ch;
     uint16_t ret;
 
+    if (size != 2) {
+        return omap_badwidth_read16(opaque, addr);
+    }
+
     switch (addr) {
     case 0x300 ... 0x3fe:
         if (s->model <= omap_dma_3_1 || !s->omap_3_1_mapping_disabled) {
@@ -1614,11 +1608,15 @@ static uint32_t omap_dma_read(void *opaque, target_phys_addr_t addr)
 }
 
 static void omap_dma_write(void *opaque, target_phys_addr_t addr,
-                uint32_t value)
+                           uint64_t value, unsigned size)
 {
     struct omap_dma_s *s = (struct omap_dma_s *) opaque;
     int reg, ch;
 
+    if (size != 2) {
+        return omap_badwidth_write16(opaque, addr, value);
+    }
+
     switch (addr) {
     case 0x300 ... 0x3fe:
         if (s->model <= omap_dma_3_1 || !s->omap_3_1_mapping_disabled) {
@@ -1655,16 +1653,10 @@ static void omap_dma_write(void *opaque, target_phys_addr_t addr,
     OMAP_BAD_REG(addr);
 }
 
-static CPUReadMemoryFunc * const omap_dma_readfn[] = {
-    omap_badwidth_read16,
-    omap_dma_read,
-    omap_badwidth_read16,
-};
-
-static CPUWriteMemoryFunc * const omap_dma_writefn[] = {
-    omap_badwidth_write16,
-    omap_dma_write,
-    omap_badwidth_write16,
+static const MemoryRegionOps omap_dma_ops = {
+    .read = omap_dma_read,
+    .write = omap_dma_write,
+    .endianness = DEVICE_NATIVE_ENDIAN,
 };
 
 static void omap_dma_request(void *opaque, int drq, int req)
@@ -1752,12 +1744,13 @@ static void omap_dma_setcaps(struct omap_dma_s *s)
 }
 
 struct soc_dma_s *omap_dma_init(target_phys_addr_t base, qemu_irq *irqs,
+                MemoryRegion *sysmem,
                 qemu_irq lcd_irq, struct omap_mpu_state_s *mpu, omap_clk clk,
                 enum omap_dma_model model)
 {
-    int iomemtype, num_irqs, memsize, i;
+    int num_irqs, memsize, i;
     struct omap_dma_s *s = (struct omap_dma_s *)
-            qemu_mallocz(sizeof(struct omap_dma_s));
+            g_malloc0(sizeof(struct omap_dma_s));
 
     if (model <= omap_dma_3_1) {
         num_irqs = 6;
@@ -1795,9 +1788,8 @@ struct soc_dma_s *omap_dma_init(target_phys_addr_t base, qemu_irq *irqs,
     omap_dma_reset(s->dma);
     omap_dma_clk_update(s, 0, 1);
 
-    iomemtype = cpu_register_io_memory(omap_dma_readfn,
-                    omap_dma_writefn, s, DEVICE_NATIVE_ENDIAN);
-    cpu_register_physical_memory(base, memsize, iomemtype);
+    memory_region_init_io(&s->iomem, &omap_dma_ops, s, "omap.dma", memsize);
+    memory_region_add_subregion(sysmem, base, &s->iomem);
 
     mpu->drq = s->dma->drq;
 
@@ -1825,12 +1817,17 @@ static void omap_dma_interrupts_4_update(struct omap_dma_s *s)
         qemu_irq_raise(s->irq[3]);
 }
 
-static uint32_t omap_dma4_read(void *opaque, target_phys_addr_t addr)
+static uint64_t omap_dma4_read(void *opaque, target_phys_addr_t addr,
+                               unsigned size)
 {
     struct omap_dma_s *s = (struct omap_dma_s *) opaque;
     int irqn = 0, chnum;
     struct omap_dma_channel_s *ch;
 
+    if (size == 1) {
+        return omap_badwidth_read16(opaque, addr);
+    }
+
     switch (addr) {
     case 0x00:	/* DMA4_REVISION */
         if (cpu_is_omap3630(s->mpu)) {
@@ -1889,20 +1886,6 @@ static uint32_t omap_dma4_read(void *opaque, target_phys_addr_t addr)
     /* Per-channel registers */
     switch (addr) {
     case 0x00:	/* DMA4_CCR */
-        TRACE("CCR = 0x%08x",  (ch->buf_disable << 25) |
-              (ch->src_sync << 24) |
-              (ch->prefetch << 23) |
-              ((ch->sync & 0x60) << 14) |
-              (ch->bs << 18) |
-              (ch->transparent_copy << 17) |
-              (ch->constant_fill << 16) |
-              (ch->mode[1] << 14) |
-              (ch->mode[0] << 12) |
-              (0 << 10) | (0 << 9) |
-              (ch->suspend << 8) |
-              (ch->enable << 7) |
-              (ch->priority << 6) |
-              (ch->fs << 5) | (ch->sync & 0x1f));
         return (ch->buf_disable << 25) |
                 (ch->src_sync << 24) |
                 (ch->prefetch << 23) |
@@ -1925,7 +1908,6 @@ static uint32_t omap_dma4_read(void *opaque, target_phys_addr_t addr)
         return ch->interrupts;
 
     case 0x0c:	/* DMA4_CSR */
-        TRACE("CSR = 0x%04x", ch->cstatus);
         return ch->cstatus;
 
     case 0x10:	/* DMA4_CSDP */
@@ -1984,21 +1966,18 @@ static uint32_t omap_dma4_read(void *opaque, target_phys_addr_t addr)
 
     case 0x50: /* DMA4_CDP */
         if (cpu_is_omap3630(s->mpu)) {
-            TRACE("CDP = 0x%08x", ch->sgl.ptr);
             return ch->sgl.ptr;
         }
         break;
 
     case 0x54: /* DMA4_CNDP */
         if (cpu_is_omap3630(s->mpu)) {
-            TRACE("CNDP = 0x%08x", ch->sgl.next);
             return ch->sgl.next;
         }
         break;
 
     case 0x58: /* DMA4_CCDN */
         if (cpu_is_omap3630(s->mpu)) {
-            TRACE("CCDN = 0x%08x", ch->sgl.number);
             return ch->sgl.number;
         }
         break;
@@ -2006,7 +1985,7 @@ static uint32_t omap_dma4_read(void *opaque, target_phys_addr_t addr)
     default:
         break;
     }
-    
+
     OMAP_BAD_REG(0x80 + chnum * 0x60 + addr);
     return 0;
 }
@@ -2021,7 +2000,7 @@ static void omap_dma4_write_ch(struct omap_dma_s *s,
         ch->src_sync = (value >> 24) & 1;	/* XXX For CamDMA must be 1 */
         if (ch->buf_disable && !ch->src_sync)
             fprintf(stderr, "%s: Buffering disable is not allowed in "
-                    "destination synchronised mode\n", __FUNCTION__);
+                            "destination synchronised mode\n", __FUNCTION__);
         ch->prefetch = (value >> 23) & 1;
         ch->bs = (value >> 18) & 1;
         ch->transparent_copy = (value >> 17) & 1;
@@ -2033,19 +2012,15 @@ static void omap_dma4_write_ch(struct omap_dma_s *s,
         ch->fs = (value & 0x0020) >> 5;
         if (ch->fs && ch->bs && ch->mode[0] && ch->mode[1])
             fprintf(stderr, "%s: For a packet transfer at least one port "
-                    "must be constant-addressed\n", __FUNCTION__);
+                            "must be constant-addressed\n", __FUNCTION__);
         ch->sync = (value & 0x001f) | ((value >> 14) & 0x0060);
         /* XXX must be 0x01 for CamDMA */
 
-        TRACE("CCR=0x%08x, transfer mode = %d, fast = %d", value,
-              (ch->sgl.ptr >> 8) & 3, !!(ch->sgl.ptr & (1 << 10)));
         if (value & 0x0080) {
             if (((ch->sgl.ptr >> 8) & 3) != 1 || /* TRANSFER_MODE */
                 !(ch->sgl.ptr & (1 << 10)) ||    /* FAST */
                 omap_dma_sgl_next(s, ch)) {
                 omap_dma_enable_channel(s, ch);
-            } else {
-                TRACE("failed to enable dma channel");
             }
         } else {
             omap_dma_disable_channel(s, ch);
@@ -2065,12 +2040,10 @@ static void omap_dma4_write_ch(struct omap_dma_s *s,
         } else {
             ch->interrupts = value & 0x09be;
         }
-        TRACE("CICR = 0x%04x", ch->interrupts);
         break;
 
     case 0x0c:	/* DMA4_CSR */
         ch->cstatus &= ~value;
-        TRACE("CSR = 0x%04x --> 0x%04x", value, ch->cstatus);
         break;
 
     case 0x10:	/* DMA4_CSDP */
@@ -2080,7 +2053,7 @@ static void omap_dma4_write_ch(struct omap_dma_s *s,
         ch->endian_lock[1] =(value >> 18) & 1;
         if (ch->endian[0] != ch->endian[1])
             fprintf(stderr, "%s: DMA endiannes conversion enable attempt\n",
-                    __FUNCTION__);
+                            __FUNCTION__);
         ch->write_mode = (value >> 16) & 3;
         ch->burst[1] = (value & 0xc000) >> 14;
         ch->pack[1] = (value & 0x2000) >> 13;
@@ -2090,7 +2063,7 @@ static void omap_dma4_write_ch(struct omap_dma_s *s,
         ch->translate[0] = (value & 0x003c) >> 2;
         if (ch->translate[0] | ch->translate[1])
             fprintf(stderr, "%s: bad MReqAddressTranslate sideband signal\n",
-                    __FUNCTION__);
+                            __FUNCTION__);
         ch->data_type = 1 << (value & 3);
         if ((value & 3) == 3)
             printf("%s: bad data_type for DMA channel\n", __FUNCTION__);
@@ -2099,15 +2072,11 @@ static void omap_dma4_write_ch(struct omap_dma_s *s,
     case 0x14:	/* DMA4_CCEN */
         ch->set_update = 1;
         ch->elements = value & 0xffffff;
-        TRACE("elements=%d, frames=%d, data=%d bytes",
-              ch->elements, ch->frames, ch->data_type);
         break;
 
     case 0x18:	/* DMA4_CFN */
         ch->frames = value & 0xffff;
         ch->set_update = 1;
-        TRACE("elements=%d, frames=%d, data=%d bytes",
-              ch->elements, ch->frames, ch->data_type);
         break;
 
     case 0x1c:	/* DMA4_CSSA */
@@ -2147,82 +2116,86 @@ static void omap_dma4_write_ch(struct omap_dma_s *s,
         /* ignore */
         break;
 
-    case 0x44:	/* DMA4_COLOR */
+    case 0x44: /* DMA4_COLOR */
         /* XXX only in sDMA */
         ch->color = value;
         break;
 
     case 0x50: /* DMA4_CDP */
         if (cpu_is_omap3630(s->mpu)) {
-            TRACE("CDP = 0x%08x", value);
             ch->sgl.ptr = value & 0x7ff;
         }
         break;
 
     case 0x54: /* DMA4_CNDP */
         if (cpu_is_omap3630(s->mpu)) {
-            TRACE("CNDP = 0x%08x", value);
             ch->sgl.next = value & ~3;
         }
         break;
 
     case 0x58: /* DMA4_CCDN */
         if (cpu_is_omap3630(s->mpu)) {
-            TRACE("CCDN = 0x%08x", value);
             ch->sgl.number = value & 0xffff;
         }
         break;
 
     default:
         fprintf(stderr, "%s: unknown register 0x%x\n",
-                __FUNCTION__, offset + 0x80);
+                __func__, offset + 0x80);
         break;
     }
 }
 
 static void omap_dma4_write(void *opaque, target_phys_addr_t addr,
-                uint32_t value)
+                            uint64_t value, unsigned size)
 {
     struct omap_dma_s *s = (struct omap_dma_s *) opaque;
     int chnum, irqn = 0;
 
+    if (size == 1) {
+        return omap_badwidth_write16(opaque, addr, value);
+    }
+
     switch (addr) {
-    case 0x14:	/* DMA4_IRQSTATUS_L3 */
-        irqn ++;
-    case 0x10:	/* DMA4_IRQSTATUS_L2 */
-        irqn ++;
-    case 0x0c:	/* DMA4_IRQSTATUS_L1 */
-        irqn ++;
-    case 0x08:	/* DMA4_IRQSTATUS_L0 */
+    case 0x14: /* DMA4_IRQSTATUS_L3 */
+        irqn++;
+    case 0x10: /* DMA4_IRQSTATUS_L2 */
+        irqn++;
+    case 0x0c: /* DMA4_IRQSTATUS_L1 */
+        irqn++;
+    case 0x08: /* DMA4_IRQSTATUS_L0 */
         s->irqstat[irqn] &= ~value;
-        if (!s->irqstat[irqn])
+        if (!s->irqstat[irqn]) {
             qemu_irq_lower(s->irq[irqn]);
+        }
         break;
 
-    case 0x24:	/* DMA4_IRQENABLE_L3 */
-        irqn ++;
-    case 0x20:	/* DMA4_IRQENABLE_L2 */
-        irqn ++;
-    case 0x1c:	/* DMA4_IRQENABLE_L1 */
-        irqn ++;
-    case 0x18:	/* DMA4_IRQENABLE_L0 */
+    case 0x24: /* DMA4_IRQENABLE_L3 */
+        irqn++;
+    case 0x20: /* DMA4_IRQENABLE_L2 */
+        irqn++;
+    case 0x1c: /* DMA4_IRQENABLE_L1 */
+        irqn++;
+    case 0x18: /* DMA4_IRQENABLE_L0 */
         s->irqen[irqn] = value;
         break;
 
-    case 0x2c:	/* DMA4_OCP_SYSCONFIG */
+    case 0x2c: /* DMA4_OCP_SYSCONFIG */
         if (value & 2) { /* SOFTRESET */
             /* N/A on 3630GP?? */
             omap_dma_reset(s->dma);
         }
         s->ocp = value & 0x3321;
-        if (((s->ocp >> 12) & 3) == 3)				/* MIDLEMODE */
-            fprintf(stderr, "%s: invalid DMA power mode\n", __FUNCTION__);
+        if (((s->ocp >> 12) & 3) == 3) { /* MIDLEMODE */
+            fprintf(stderr, "%s: invalid DMA power mode\n", __func__);
+        }
         break;
 
-    case 0x78:	/* DMA4_GCR */
+    case 0x78: /* DMA4_GCR */
         s->gcr = value & 0x00ff00ff;
-        if ((value & 0xff) == 0x00)		/* MAX_CHANNEL_FIFO_DEPTH */
-            fprintf(stderr, "%s: wrong FIFO depth in GCR\n", __FUNCTION__);
+        if ((value & 0xff) == 0x00) { /* MAX_CHANNEL_FIFO_DEPTH */
+            fprintf(stderr, "%s: wrong FIFO depth in GCR\n", __func__);
+        }
         break;
 
     case 0x80 ... 0xfff:
@@ -2232,12 +2205,12 @@ static void omap_dma4_write(void *opaque, target_phys_addr_t addr,
         omap_dma4_write_ch(s, s->ch + chnum, (uint32_t)addr, value);
         break;
 
-    case 0x00:	/* DMA4_REVISION */
-    case 0x28:	/* DMA4_SYSSTATUS */
-    case 0x64:	/* DMA4_CAPS_0 */
-    case 0x6c:	/* DMA4_CAPS_2 */
-    case 0x70:	/* DMA4_CAPS_3 */
-    case 0x74:	/* DMA4_CAPS_4 */
+    case 0x00: /* DMA4_REVISION */
+    case 0x28: /* DMA4_SYSSTATUS */
+    case 0x64: /* DMA4_CAPS_0 */
+    case 0x6c: /* DMA4_CAPS_2 */
+    case 0x70: /* DMA4_CAPS_3 */
+    case 0x74: /* DMA4_CAPS_4 */
         OMAP_RO_REG(addr);
         break;
 
@@ -2247,16 +2220,10 @@ static void omap_dma4_write(void *opaque, target_phys_addr_t addr,
     }
 }
 
-static CPUReadMemoryFunc * const omap_dma4_readfn[] = {
-    omap_badwidth_read16,
-    omap_dma4_read,
-    omap_dma4_read,
-};
-
-static CPUWriteMemoryFunc * const omap_dma4_writefn[] = {
-    omap_badwidth_write16,
-    omap_dma4_write,
-    omap_dma4_write,
+static const MemoryRegionOps omap_dma4_ops = {
+    .read = omap_dma4_read,
+    .write = omap_dma4_write,
+    .endianness = DEVICE_NATIVE_ENDIAN,
 };
 
 static struct omap_dma_s *omap_dma4_init_internal(struct omap_mpu_state_s *mpu,
@@ -2266,8 +2233,8 @@ static struct omap_dma_s *omap_dma4_init_internal(struct omap_mpu_state_s *mpu,
 {
     int i;
     struct omap_dma_s *s = (struct omap_dma_s *)
-        qemu_mallocz(sizeof(struct omap_dma_s));
-    
+            g_malloc0(sizeof(struct omap_dma_s));
+
     s->model = omap_dma_4;
     s->chans = chans;
     s->mpu = mpu;
@@ -2296,32 +2263,30 @@ static struct omap_dma_s *omap_dma4_init_internal(struct omap_mpu_state_s *mpu,
 
     return s;
 }
-    
+
 struct soc_dma_s *omap_dma4_init(target_phys_addr_t base, qemu_irq *irqs,
+                MemoryRegion *sysmem,
                 struct omap_mpu_state_s *mpu, int fifo,
                 int chans, omap_clk iclk, omap_clk fclk)
 {
-    int iomemtype;
     struct omap_dma_s *s = omap_dma4_init_internal(mpu, irqs, chans, 64,
                                                    iclk, fclk);
+    memory_region_init_io(&s->iomem, &omap_dma4_ops, s, "omap.dma4", 0x1000);
+    memory_region_add_subregion(sysmem, base, &s->iomem);
 
-    iomemtype = cpu_register_io_memory(omap_dma4_readfn,
-                    omap_dma4_writefn, s, DEVICE_NATIVE_ENDIAN);
-    cpu_register_physical_memory(base, 0x1000, iomemtype);
-    
     return s->dma;
 }
 
 struct soc_dma_s *omap3_dma4_init(struct omap_target_agent_s *ta,
                                   struct omap_mpu_state_s *mpu,
-                                  qemu_irq *irqs, int chans,
+                                  qemu_irq *irqs,
+                                  int chans,
                                   omap_clk iclk, omap_clk fclk)
 {
     struct omap_dma_s *s = omap_dma4_init_internal(mpu, irqs, chans, 96,
                                                    iclk, fclk);
-    omap_l4_attach(ta, 0, cpu_register_io_memory(omap_dma4_readfn,
-                                                 omap_dma4_writefn, s,
-                                                 DEVICE_NATIVE_ENDIAN));
+    memory_region_init_io(&s->iomem, &omap_dma4_ops, s, "omap.dma4", 0x1000);
+    omap_l4_attach(ta, 0, &s->iomem);
     return s->dma;
 }
 
