@@ -128,6 +128,8 @@ struct omap_dss_panel_s {
 
 struct omap_dss_s {
     SysBusDevice busdev;
+    MemoryRegion iomem_diss1, iomem_disc1, iomem_rfbi1, iomem_venc1, iomem_im3;
+    MemoryRegion iomem_dsi;
     int32_t mpu_model;
     qemu_irq irq;
     qemu_irq drq;
@@ -375,7 +377,7 @@ static void omap_rfbi_transfer_start(struct omap_dss_s *s)
     }
     if (!data) {
         if (len > bounce_len) {
-            bounce_buffer = qemu_realloc(bounce_buffer, len);
+            bounce_buffer = g_realloc(bounce_buffer, len);
         }
         data = bounce_buffer;
         cpu_physical_memory_read(data_addr, data, len);
@@ -459,7 +461,7 @@ static void omap_dsi_transfer_start(struct omap_dss_s *s, int ch)
     }
 }
 
-static void omap_dss_panel_layer_update(DisplayState *ds,
+static void omap_dss_panel_layer_update(DisplayState *ds, MemoryRegion *mr,
                                         uint32_t panel_width,
                                         uint32_t panel_height,
                                         uint32_t posx,
@@ -495,7 +497,7 @@ static void omap_dss_panel_layer_update(DisplayState *ds,
     uint32_t copy_height = ((*posy) + height) > panel_height
                            ? (panel_height - (*posy)) : height;
     uint32_t linesize = ds_get_linesize(ds);
-    framebuffer_update_display(ds, addr, copy_width, copy_height,
+    framebuffer_update_display(ds, mr, addr, copy_width, copy_height,
                                (format < 3)
                                ? (width >> (3 - format))
                                : (width * omap_lcd_Bpp[format]),
@@ -504,7 +506,8 @@ static void omap_dss_panel_layer_update(DisplayState *ds,
                                posy, endy);
 }
 
-static void omap_dss_panel_update_display(struct omap_dss_panel_s *s, int lcd)
+static void omap_dss_panel_update_display(struct omap_dss_panel_s *s,
+                                          MemoryRegion *mr, int lcd)
 {
     if (s->invalidate) {
         if (s->shadow.width != ds_get_width(s->ds) 
@@ -528,7 +531,8 @@ static void omap_dss_panel_update_display(struct omap_dss_panel_s *s, int lcd)
                                      * sizeof(uint32_t));
         }
         first_row = s->shadow.gfx.posy;
-        omap_dss_panel_layer_update(s->ds, s->shadow.width, s->shadow.height,
+        omap_dss_panel_layer_update(s->ds, mr,
+                                    s->shadow.width, s->shadow.height,
                                     s->shadow.gfx.posx, &first_row, &last_row,
                                     s->shadow.gfx.nx, s->shadow.gfx.ny,
                                     s->shadow.gfx.attr, s->shadow.gfx.addr[0],
@@ -551,11 +555,11 @@ static void omap_lcd_panel_update_display(void *opaque)
         || (s->lcd.shadow.control & (1 << 11))) { /* STALLMODE */
         return;
     }
-    omap_dss_panel_update_display(&s->lcd, 1);
+    omap_dss_panel_update_display(&s->lcd, sysbus_address_space(&s->busdev), 1);
     omap_dss_framedone(s);
 }
 
-void omap_lcd_panel_invalidate_display(void *opaque)
+static void omap_lcd_panel_invalidate_display(void *opaque)
 {
     struct omap_dss_s *s = opaque;
     s->lcd.invalidate = 1;
@@ -567,7 +571,7 @@ static void omap_dig_panel_update_display(void *opaque)
     if (!s->dig.ds || !(s->dig.shadow.control & 2)) { /* DIGITALENABLE */
         return;
     }
-    omap_dss_panel_update_display(&s->dig, 0);
+    omap_dss_panel_update_display(&s->dig, sysbus_address_space(&s->busdev), 0);
     omap_dss_framedone(s);
 }
 
@@ -599,12 +603,11 @@ static void omap_dss_panel_go(struct omap_dss_s *s,
         }
         if (new_size != p->shadow.gfx_palette_size) {
             if (p->shadow.gfx_palette) {
-                qemu_free(p->shadow.gfx_palette);
+                g_free(p->shadow.gfx_palette);
                 p->shadow.gfx_palette = NULL;
             }
             if (new_size) {
-                p->shadow.gfx_palette = qemu_malloc(new_size *
-                                                    sizeof(uint32_t));
+                p->shadow.gfx_palette = g_malloc(new_size * sizeof(uint32_t));
             }
             p->shadow.gfx_palette_size = new_size;
         }
@@ -620,7 +623,7 @@ static void omap_dss_panel_reset(void *opaque)
         memset(&s->shadow.vid1, 0, sizeof(s->shadow.vid1));
         memset(&s->shadow.vid2, 0, sizeof(s->shadow.vid2));
         if (s->shadow.gfx_palette) {
-            qemu_free(s->shadow.gfx_palette);
+            g_free(s->shadow.gfx_palette);
             s->shadow.gfx_palette = NULL;
             s->shadow.gfx_palette_size = 0;
         }
@@ -779,11 +782,16 @@ static void omap_dss_reset(DeviceState *dev)
     omap_dss_interrupt_update(s);
 }
 
-static uint32_t omap_diss_read(void *opaque, target_phys_addr_t addr)
+static uint64_t omap_diss_read(void *opaque, target_phys_addr_t addr,
+                               unsigned size)
 {
     struct omap_dss_s *s = (struct omap_dss_s *) opaque;
     uint32_t x;
 
+    if (size != 4) {
+        return omap_badwidth_read32(opaque, addr);
+    }
+
     switch (addr) {
     case 0x00:	/* DSS_REVISIONNUMBER */
         TRACEDISS("DSS_REVISIONNUMBER: 0x20");
@@ -858,10 +866,14 @@ static uint32_t omap_diss_read(void *opaque, target_phys_addr_t addr)
 }
 
 static void omap_diss_write(void *opaque, target_phys_addr_t addr,
-                uint32_t value)
+                            uint64_t value, unsigned size)
 {
     struct omap_dss_s *s = (struct omap_dss_s *) opaque;
 
+    if (size != 4) {
+        return omap_badwidth_write32(opaque, addr, value);
+    }
+
     switch (addr) {
     case 0x00:	/* DSS_REVISIONNUMBER */
     case 0x14:	/* DSS_SYSSTATUS */
@@ -937,23 +949,22 @@ static void omap_diss_write(void *opaque, target_phys_addr_t addr,
     }
 }
 
-static CPUReadMemoryFunc * const omap_diss1_readfn[] = {
-    omap_badwidth_read32,
-    omap_badwidth_read32,
-    omap_diss_read,
+static const MemoryRegionOps omap_diss_ops = {
+    .read = omap_diss_read,
+    .write = omap_diss_write,
+    .endianness = DEVICE_NATIVE_ENDIAN,
 };
 
-static CPUWriteMemoryFunc * const omap_diss1_writefn[] = {
-    omap_badwidth_write32,
-    omap_badwidth_write32,
-    omap_diss_write,
-};
-
-static uint32_t omap_disc_read(void *opaque, target_phys_addr_t addr)
+static uint64_t omap_disc_read(void *opaque, target_phys_addr_t addr,
+                               unsigned size)
 {
     struct omap_dss_s *s = (struct omap_dss_s *) opaque;
     int n = 0;
 
+    if (size != 4) {
+        return omap_badwidth_read32(opaque, addr);
+    }
+
     switch (addr) {
     case 0x000:	/* DISPC_REVISION */
         TRACEDISPC("DISPC_REVISION: 0x%08x", s->dispc.rev);
@@ -1188,11 +1199,15 @@ static uint32_t omap_disc_read(void *opaque, target_phys_addr_t addr)
 }
 
 static void omap_disc_write(void *opaque, target_phys_addr_t addr,
-                uint32_t value)
+                            uint64_t value, unsigned size)
 {
     struct omap_dss_s *s = (struct omap_dss_s *) opaque;
     uint32_t n = 0;
 
+    if (size != 4) {
+        return omap_badwidth_write32(opaque, addr, value);
+    }
+
     switch (addr) {
     case 0x000: /* DISPC_REVISION */
     case 0x014: /* DISPC_SYSSTATUS */
@@ -1229,7 +1244,7 @@ static void omap_disc_write(void *opaque, target_phys_addr_t addr,
                 | (s->dispc.control & 0x6000);
         }
         if (value & (1 << 12))			/* OVERLAY_OPTIMIZATION */
-            if (~((s->dispc.plane[1].attr | s->dispc.plane[2].attr) & 1)) {
+            if (!((s->dispc.plane[1].attr | s->dispc.plane[2].attr) & 1)) {
                  TRACEDISPC("Overlay Optimization when no overlay "
                             "region effectively exists leads to "
                             "unpredictable behaviour!");
@@ -1513,22 +1528,21 @@ static void omap_disc_write(void *opaque, target_phys_addr_t addr,
     }
 }
 
-static CPUReadMemoryFunc * const omap_disc1_readfn[] = {
-    omap_badwidth_read32,
-    omap_badwidth_read32,
-    omap_disc_read,
+static const MemoryRegionOps omap_disc_ops = {
+    .read = omap_disc_read,
+    .write = omap_disc_write,
+    .endianness = DEVICE_NATIVE_ENDIAN,
 };
 
-static CPUWriteMemoryFunc * const omap_disc1_writefn[] = {
-    omap_badwidth_write32,
-    omap_badwidth_write32,
-    omap_disc_write,
-};
-
-static uint32_t omap_rfbi_read(void *opaque, target_phys_addr_t addr)
+static uint64_t omap_rfbi_read(void *opaque, target_phys_addr_t addr,
+                               unsigned size)
 {
     struct omap_dss_s *s = (struct omap_dss_s *) opaque;
 
+    if (size != 4) {
+        return omap_badwidth_read32(opaque, addr);
+    }
+
     switch (addr) {
     case 0x00:	/* RFBI_REVISION */
         TRACERFBI("RFBI_REVISION: 0x10");
@@ -1609,10 +1623,14 @@ static uint32_t omap_rfbi_read(void *opaque, target_phys_addr_t addr)
 }
 
 static void omap_rfbi_write(void *opaque, target_phys_addr_t addr,
-                uint32_t value)
+                            uint64_t value, unsigned size)
 {
     struct omap_dss_s *s = (struct omap_dss_s *) opaque;
 
+    if (size != 4) {
+        return omap_badwidth_write32(opaque, addr, value);
+    }
+
     switch (addr) {
     case 0x10:	/* RFBI_SYSCONFIG */
         TRACERFBI("RFBI_SYSCONFIG = 0x%08x", value);
@@ -1679,7 +1697,7 @@ static void omap_rfbi_write(void *opaque, target_phys_addr_t addr,
         if ((s->rfbi.control & (1 << 2)) && s->rfbi.chip[0])
             s->rfbi.rxbuf = s->rfbi.chip[0]->read(s->rfbi.chip[0]->opaque, 1);
         else if ((s->rfbi.control & (1 << 3)) && s->rfbi.chip[1])
-            s->rfbi.rxbuf = s->rfbi.chip[0]->read(s->rfbi.chip[0]->opaque, 1);
+            s->rfbi.rxbuf = s->rfbi.chip[1]->read(s->rfbi.chip[1]->opaque, 1);
         if (!-- s->rfbi.pixels)
             omap_rfbi_transfer_stop(s);
         break;
@@ -1689,7 +1707,7 @@ static void omap_rfbi_write(void *opaque, target_phys_addr_t addr,
         if ((s->rfbi.control & (1 << 2)) && s->rfbi.chip[0])
             s->rfbi.rxbuf = s->rfbi.chip[0]->read(s->rfbi.chip[0]->opaque, 0);
         else if ((s->rfbi.control & (1 << 3)) && s->rfbi.chip[1])
-            s->rfbi.rxbuf = s->rfbi.chip[0]->read(s->rfbi.chip[0]->opaque, 0);
+            s->rfbi.rxbuf = s->rfbi.chip[1]->read(s->rfbi.chip[1]->opaque, 0);
         if (!-- s->rfbi.pixels)
             omap_rfbi_transfer_stop(s);
         break;
@@ -1760,20 +1778,19 @@ static void omap_rfbi_write(void *opaque, target_phys_addr_t addr,
     }
 }
 
-static CPUReadMemoryFunc * const omap_rfbi1_readfn[] = {
-    omap_badwidth_read32,
-    omap_badwidth_read32,
-    omap_rfbi_read,
+static const MemoryRegionOps omap_rfbi_ops = {
+    .read = omap_rfbi_read,
+    .write = omap_rfbi_write,
+    .endianness = DEVICE_NATIVE_ENDIAN,
 };
 
-static CPUWriteMemoryFunc * const omap_rfbi1_writefn[] = {
-    omap_badwidth_write32,
-    omap_badwidth_write32,
-    omap_rfbi_write,
-};
-
-static uint32_t omap_venc_read(void *opaque, target_phys_addr_t addr)
+static uint64_t omap_venc_read(void *opaque, target_phys_addr_t addr,
+                               unsigned size)
 {
+    if (size != 4) {
+        return omap_badwidth_read32(opaque, addr);
+    }
+
     switch (addr) {
     case 0x00:	/* REV_ID */
         return 0x2;
@@ -1828,8 +1845,12 @@ static uint32_t omap_venc_read(void *opaque, target_phys_addr_t addr)
 }
 
 static void omap_venc_write(void *opaque, target_phys_addr_t addr,
-                uint32_t value)
+                            uint64_t value, unsigned size)
 {
+    if (size != 4) {
+        return omap_badwidth_write32(opaque, addr, size);
+    }
+
     switch (addr) {
     case 0x00: /* REV_ID */
     case 0x04: /* STATUS */
@@ -1883,20 +1904,19 @@ static void omap_venc_write(void *opaque, target_phys_addr_t addr,
     }
 }
 
-static CPUReadMemoryFunc * const omap_venc1_readfn[] = {
-    omap_badwidth_read32,
-    omap_badwidth_read32,
-    omap_venc_read,
+static const MemoryRegionOps omap_venc_ops = {
+    .read = omap_venc_read,
+    .write = omap_venc_write,
+    .endianness = DEVICE_NATIVE_ENDIAN,
 };
 
-static CPUWriteMemoryFunc * const omap_venc1_writefn[] = {
-    omap_badwidth_write32,
-    omap_badwidth_write32,
-    omap_venc_write,
-};
-
-static uint32_t omap_im3_read(void *opaque, target_phys_addr_t addr)
+static uint64_t omap_im3_read(void *opaque, target_phys_addr_t addr,
+                              unsigned size)
 {
+    if (size != 4) {
+        return omap_badwidth_read32(opaque, addr);
+    }
+
     switch (addr) {
     case 0x0a8:	/* SBIMERRLOGA */
     case 0x0b0:	/* SBIMERRLOG */
@@ -1917,8 +1937,12 @@ static uint32_t omap_im3_read(void *opaque, target_phys_addr_t addr)
 }
 
 static void omap_im3_write(void *opaque, target_phys_addr_t addr,
-                uint32_t value)
+                           uint64_t value, unsigned size)
 {
+    if (size != 4) {
+        return omap_badwidth_write32(opaque, addr, value);
+    }
+
     switch (addr) {
     case 0x0b0:	/* SBIMERRLOG */
     case 0x190:	/* SBIMSTATE */
@@ -1934,16 +1958,10 @@ static void omap_im3_write(void *opaque, target_phys_addr_t addr,
     }
 }
 
-static CPUReadMemoryFunc * const omap_im3_readfn[] = {
-    omap_badwidth_read32,
-    omap_badwidth_read32,
-    omap_im3_read,
-};
-
-static CPUWriteMemoryFunc * const omap_im3_writefn[] = {
-    omap_badwidth_write32,
-    omap_badwidth_write32,
-    omap_im3_write,
+static const MemoryRegionOps omap_im3_ops = {
+    .read = omap_im3_read,
+    .write = omap_im3_write,
+    .endianness = DEVICE_NATIVE_ENDIAN,
 };
 
 static void omap_dsi_push_rx_fifo(struct omap_dss_s *s, int ch, uint32_t value)
@@ -2446,16 +2464,20 @@ static void omap_dsi_write(void *opaque, target_phys_addr_t addr,
     }
 }
 
-static CPUReadMemoryFunc *omap_dsi_readfn[] = {
-    omap_badwidth_read32,
-    omap_badwidth_read32,
-    omap_dsi_read,
-};
-
-static CPUWriteMemoryFunc *omap_dsi_writefn[] = {
-    omap_badwidth_write32,
-    omap_badwidth_write32,
-    omap_dsi_write,
+static const MemoryRegionOps omap_dsi_ops = {
+    .old_mmio = {
+        .read = {
+            omap_badwidth_read32,
+            omap_badwidth_read32,
+            omap_dsi_read,
+        },
+        .write = {
+            omap_badwidth_write32,
+            omap_badwidth_write32,
+            omap_dsi_write,
+        },
+    },
+    .endianness = DEVICE_NATIVE_ENDIAN,
 };
 
 static int omap_dss_init(SysBusDevice *dev)
@@ -2463,30 +2485,27 @@ static int omap_dss_init(SysBusDevice *dev)
     struct omap_dss_s *s = FROM_SYSBUS(struct omap_dss_s, dev);
     sysbus_init_irq(dev, &s->irq);
     sysbus_init_irq(dev, &s->drq); /* linetrigger */
-    sysbus_init_mmio(dev, 0x400,
-                     cpu_register_io_memory(omap_diss1_readfn,
-                                            omap_diss1_writefn, s,
-                                            DEVICE_NATIVE_ENDIAN));
-    sysbus_init_mmio(dev, 0x400,
-                     cpu_register_io_memory(omap_disc1_readfn,
-                                            omap_disc1_writefn, s,
-                                            DEVICE_NATIVE_ENDIAN));
-    sysbus_init_mmio(dev, 0x400,
-                     cpu_register_io_memory(omap_rfbi1_readfn,
-                                            omap_rfbi1_writefn, s,
-                                            DEVICE_NATIVE_ENDIAN));
-    sysbus_init_mmio(dev, 0x400,
-                     cpu_register_io_memory(omap_venc1_readfn,
-                                            omap_venc1_writefn, s,
-                                            DEVICE_NATIVE_ENDIAN));
+
+    memory_region_init_io(&s->iomem_diss1, &omap_diss_ops, s,
+                          "omap.diss1", 0x400);
+    memory_region_init_io(&s->iomem_disc1, &omap_disc_ops, s,
+                          "omap.disc1", 0x400);
+    memory_region_init_io(&s->iomem_rfbi1, &omap_rfbi_ops, s,
+                          "omap.rfbi1", 0x400);
+    memory_region_init_io(&s->iomem_venc1, &omap_venc_ops, s,
+                          "omap.venc1", 0x400);
+    sysbus_init_mmio(dev, &s->iomem_diss1);
+    sysbus_init_mmio(dev, &s->iomem_disc1);
+    sysbus_init_mmio(dev, &s->iomem_rfbi1);
+    sysbus_init_mmio(dev, &s->iomem_venc1);
+
     if (s->mpu_model < omap2410) {
         hw_error("%s: unsupported cpu type\n", __FUNCTION__);
     } else if (s->mpu_model < omap3430) {
         s->dispc.rev = 0x20;
-        sysbus_init_mmio(dev, 0x1000,
-                         cpu_register_io_memory(omap_im3_readfn,
-                                                omap_im3_writefn, s,
-                                                DEVICE_NATIVE_ENDIAN));
+        memory_region_init_io(&s->iomem_im3, &omap_im3_ops, s,
+                              "omap.im3", 0x1000);
+        sysbus_init_mmio(dev, &s->iomem_im3);
     } else {
         s->dispc.rev = 0x30;
         s->dispc.lcdframer = qemu_new_timer_ns(vm_clock, omap_dss_framedone, s);
@@ -2495,10 +2514,9 @@ static int omap_dss_init(SysBusDevice *dev)
                                     omap_dss_linefn);
         s->dsi.xfer_timer = qemu_new_timer_ns(vm_clock, omap_dsi_transfer_stop,
                                            s);
-        sysbus_init_mmio(dev, 0x400,
-                         cpu_register_io_memory(omap_dsi_readfn,
-                                                omap_dsi_writefn, s,
-                                                DEVICE_NATIVE_ENDIAN));
+        memory_region_init_io(&s->iomem_dsi, &omap_dsi_ops, s,
+                              "omap.dsi", 0x400);
+        sysbus_init_mmio(dev, &s->iomem_dsi);
         sysbus_init_irq(dev, &s->dsi.drq[0]);
         sysbus_init_irq(dev, &s->dsi.drq[1]);
         sysbus_init_irq(dev, &s->dsi.drq[2]);
@@ -2507,20 +2525,30 @@ static int omap_dss_init(SysBusDevice *dev)
     return 0;
 }
 
-static SysBusDeviceInfo omap_dss_info = {
-    .init = omap_dss_init,
-    .qdev.name = "omap_dss",
-    .qdev.size = sizeof(struct omap_dss_s),
-    .qdev.reset = omap_dss_reset,
-    .qdev.props = (Property[]) {
-        DEFINE_PROP_INT32("mpu_model", struct omap_dss_s, mpu_model, 0),
-        DEFINE_PROP_END_OF_LIST()
-    }
+static Property omap_dss_properties[] = {
+    DEFINE_PROP_INT32("mpu_model", struct omap_dss_s, mpu_model, 0),
+    DEFINE_PROP_END_OF_LIST()
 };
 
-static void omap_dss_register_device(void)
+static void omap_dss_class_init(ObjectClass *klass, void *data)
 {
-    sysbus_register_withprop(&omap_dss_info);
+    DeviceClass *dc = DEVICE_CLASS(klass);
+    SysBusDeviceClass *k = SYS_BUS_DEVICE_CLASS(klass);
+    k->init = omap_dss_init;
+    dc->reset = omap_dss_reset;
+    dc->props = omap_dss_properties;
+}
+
+static TypeInfo omap_dss_info = {
+    .name = "omap_dss",
+    .parent = TYPE_SYS_BUS_DEVICE,
+    .instance_size = sizeof(struct omap_dss_s),
+    .class_init = omap_dss_class_init,
+};
+
+static void omap_dss_register_types(void)
+{
+    type_register_static(&omap_dss_info);
 }
 
 void omap_rfbi_attach(DeviceState *dev, int cs,
@@ -2570,4 +2598,4 @@ void omap_digital_panel_attach(DeviceState *dev)
     }
 }
 
-device_init(omap_dss_register_device)
+type_init(omap_dss_register_types)
