@@ -18,10 +18,12 @@
  * with this program; if not, see <http://www.gnu.org/licenses/>.
  */
 #include "hw.h"
+#include "qemu-timer.h"
 #include "omap.h"
 
 /* GP timers */
 struct omap_gp_timer_s {
+    MemoryRegion iomem;
     qemu_irq irq;
     qemu_irq wkup;
     qemu_irq in;
@@ -140,7 +142,7 @@ static inline void omap_gp_timer_update(struct omap_gp_timer_s *timer)
 
     if (timer->st && timer->rate) {
         if (timer->ticks_per_sec > 0xffffffff) {
-            rate = timer->rate >> (timer->pre ? timer->ptv + 1 : 0); /*1s -> rate ticks*/
+            rate = timer->rate >> (timer->pre ? timer->ptv + 1 : 0);
             expires = muldiv64(0x100000000ll - timer->val,
                                get_ticks_per_sec(), rate);
         } else {
@@ -151,7 +153,7 @@ static inline void omap_gp_timer_update(struct omap_gp_timer_s *timer)
 
         if (timer->ce && timer->match_val >= timer->val) {
             if (timer->ticks_per_sec > 0xffffffff) {
-                rate = timer->rate >> (timer->pre ? timer->ptv + 1 : 0); /*1s -> rate ticks*/
+                rate = timer->rate >> (timer->pre ? timer->ptv + 1 : 0);
                 matches = muldiv64(timer->match_val - timer->val,
                                    get_ticks_per_sec(), rate);
             } else {
@@ -171,14 +173,13 @@ static inline void omap_gp_timer_update(struct omap_gp_timer_s *timer)
 
 static inline void omap_gp_timer_trigger(struct omap_gp_timer_s *timer)
 {
-    if (timer->pt) {
+    if (timer->pt)
         /* TODO in overflow-and-match mode if the first event to
          * occur is the match, don't toggle.  */
         omap_gp_timer_out(timer, !timer->out_val);
-    } else {
+    else
         /* TODO inverted pulse on timer->out_val == 1?  */
         qemu_irq_pulse(timer->out);
-    }
 }
 
 static void omap_gp_timer_tick(void *opaque)
@@ -194,7 +195,7 @@ static void omap_gp_timer_tick(void *opaque)
     }
 
     if (timer->trigger == gpt_trigger_overflow ||
-        timer->trigger == gpt_trigger_both)
+                    timer->trigger == gpt_trigger_both)
         omap_gp_timer_trigger(timer);
 
     omap_gp_timer_intr(timer, GPT_OVF_IT);
@@ -253,7 +254,7 @@ static void omap_gp_timer_clk_update(void *opaque, int line, int on)
 static void omap_gp_timer_clk_setup(struct omap_gp_timer_s *timer)
 {
     omap_clk_adduser(timer->clk,
-                     qemu_allocate_irqs(omap_gp_timer_clk_update, timer, 1)[0]);
+                    qemu_allocate_irqs(omap_gp_timer_clk_update, timer, 1)[0]);
     timer->rate = omap_clk_getrate(timer->clk);
 }
 
@@ -364,58 +365,52 @@ static uint32_t omap_gp_timer_readh(void *opaque, target_phys_addr_t addr)
     }
 }
 
-static CPUReadMemoryFunc * const omap_gp_timer_readfn[] = {
-    omap_badwidth_read32,
-    omap_gp_timer_readh,
-    omap_gp_timer_readw,
-};
-
 static void omap_gp_timer_write(void *opaque, target_phys_addr_t addr,
-                                uint32_t value)
+                uint32_t value)
 {
     struct omap_gp_timer_s *s = (struct omap_gp_timer_s *) opaque;
-    
+
     switch (addr) {
-    case 0x00:  /* TIDR */
-    case 0x14:  /* TISTAT */
-    case 0x34:  /* TWPS */
-    case 0x3c:  /* TCAR1 */
-    case 0x44:  /* TCAR2 */
+    case 0x00:	/* TIDR */
+    case 0x14:	/* TISTAT */
+    case 0x34:	/* TWPS */
+    case 0x3c:	/* TCAR1 */
+    case 0x44:	/* TCAR2 */
         OMAP_RO_REG(addr);
         break;
 
-    case 0x10:  /* TIOCP_CFG */
+    case 0x10:	/* TIOCP_CFG */
         s->config = value & 0x33d;
         if (((value >> 3) & 3) == 3)				/* IDLEMODE */
             fprintf(stderr, "%s: illegal IDLEMODE value in TIOCP_CFG\n",
-                    __FUNCTION__);
+                            __FUNCTION__);
         if (value & 2)						/* SOFTRESET */
             omap_gp_timer_reset(s);
         break;
 
-    case 0x18:  /* TISR */
+    case 0x18:	/* TISR */
         if (value & GPT_TCAR_IT)
             s->capt_num = 0;
         if (s->status && !(s->status &= ~value))
             qemu_irq_lower(s->irq);
         break;
 
-    case 0x1c:  /* TIER */
+    case 0x1c:	/* TIER */
         s->it_ena = value & 7;
         break;
 
-    case 0x20:  /* TWER */
+    case 0x20:	/* TWER */
         s->wu_ena = value & 7;
         break;
 
-    case 0x24:  /* TCLR */
+    case 0x24:	/* TCLR */
         omap_gp_timer_sync(s);
         s->inout = (value >> 14) & 1;
         s->capt2 = (value >> 13) & 1;
         s->pt = (value >> 12) & 1;
         s->trigger = (value >> 10) & 3;
         if (s->capture == gpt_capture_none &&
-            ((value >> 8) & 3) != gpt_capture_none)
+                        ((value >> 8) & 3) != gpt_capture_none)
             s->capt_num = 0;
         s->capture = (value >> 8) & 3;
         s->scpwm = (value >> 7) & 1;
@@ -426,10 +421,10 @@ static void omap_gp_timer_write(void *opaque, target_phys_addr_t addr,
         s->st = (value >> 0) & 1;
         if (s->inout && s->trigger != gpt_trigger_none)
             fprintf(stderr, "%s: GP timer pin must be an output "
-                    "for this trigger mode\n", __FUNCTION__);
+                            "for this trigger mode\n", __FUNCTION__);
         if (!s->inout && s->capture != gpt_capture_none)
             fprintf(stderr, "%s: GP timer pin must be an input "
-                    "for this capture mode\n", __FUNCTION__);
+                            "for this capture mode\n", __FUNCTION__);
         if (s->trigger == gpt_trigger_none)
             omap_gp_timer_out(s, s->scpwm);
         /* TODO: make sure this doesn't overflow 32-bits */
@@ -437,29 +432,29 @@ static void omap_gp_timer_write(void *opaque, target_phys_addr_t addr,
         omap_gp_timer_update(s);
         break;
 
-    case 0x28:  /* TCRR */
+    case 0x28:	/* TCRR */
         s->time = qemu_get_clock_ns(vm_clock);
         s->val = value;
         omap_gp_timer_update(s);
         break;
 
-    case 0x2c:  /* TLDR */
+    case 0x2c:	/* TLDR */
         s->load_val = value;
         break;
 
-    case 0x30:  /* TTGR */
+    case 0x30:	/* TTGR */
         s->time = qemu_get_clock_ns(vm_clock);
         s->val = s->load_val;
         omap_gp_timer_update(s);
         break;
 
-    case 0x38:  /* TMAR */
+    case 0x38:	/* TMAR */
         omap_gp_timer_sync(s);
         s->match_val = value;
         omap_gp_timer_update(s);
         break;
 
-    case 0x40:  /* TSICR */
+    case 0x40:	/* TSICR */
         s->posted = (value >> 2) & 1;
         if (value & 2)	/* How much exactly are we supposed to reset? */
             omap_gp_timer_reset(s);
@@ -471,7 +466,7 @@ static void omap_gp_timer_write(void *opaque, target_phys_addr_t addr,
 }
 
 static void omap_gp_timer_writeh(void *opaque, target_phys_addr_t addr,
-                                 uint32_t value)
+                uint32_t value)
 {
     struct omap_gp_timer_s *s = (struct omap_gp_timer_s *) opaque;
 
@@ -481,19 +476,27 @@ static void omap_gp_timer_writeh(void *opaque, target_phys_addr_t addr,
         s->writeh = (uint16_t) value;
 }
 
-static CPUWriteMemoryFunc * const omap_gp_timer_writefn[] = {
-    omap_badwidth_write32,
-    omap_gp_timer_writeh,
-    omap_gp_timer_write,
+static const MemoryRegionOps omap_gp_timer_ops = {
+    .old_mmio = {
+        .read = {
+            omap_badwidth_read32,
+            omap_gp_timer_readh,
+            omap_gp_timer_readw,
+        },
+        .write = {
+            omap_badwidth_write32,
+            omap_gp_timer_writeh,
+            omap_gp_timer_write,
+        },
+    },
+    .endianness = DEVICE_NATIVE_ENDIAN,
 };
 
 struct omap_gp_timer_s *omap_gp_timer_init(struct omap_target_agent_s *ta,
-                                           qemu_irq irq, omap_clk fclk,
-                                           omap_clk iclk)
+                qemu_irq irq, omap_clk fclk, omap_clk iclk)
 {
-    int iomemtype;
     struct omap_gp_timer_s *s = (struct omap_gp_timer_s *)
-    qemu_mallocz(sizeof(struct omap_gp_timer_s));
+            g_malloc0(sizeof(struct omap_gp_timer_s));
 
     s->ta = ta;
     s->irq = irq;
@@ -504,9 +507,9 @@ struct omap_gp_timer_s *omap_gp_timer_init(struct omap_target_agent_s *ta,
     omap_gp_timer_reset(s);
     omap_gp_timer_clk_setup(s);
 
-    iomemtype = l4_register_io_memory(omap_gp_timer_readfn,
-                                      omap_gp_timer_writefn, s);
-    omap_l4_attach(ta, 0, iomemtype);
+    memory_region_init_io(&s->iomem, &omap_gp_timer_ops, s, "omap.gptimer",
+                          omap_l4_region_size(ta, 0));
+    omap_l4_attach(ta, 0, &s->iomem);
 
     return s;
 }
