@@ -40,6 +40,7 @@
 
 typedef struct omap_mcspi_bus_s {
     SPIBus *bus;
+    MemoryRegion iomem;
     qemu_irq irq;
     int chnum;
     uint8_t revision;
@@ -275,12 +276,17 @@ static void omap_mcspi_bus_reset(OMAPSPIBusState *s)
     omap_mcspi_interrupt_update(s);
 }
 
-static uint32_t omap_mcspi_read(void *opaque, target_phys_addr_t addr)
+static uint64_t omap_mcspi_read(void *opaque, target_phys_addr_t addr,
+                                unsigned size)
 {
     OMAPSPIBusState *s = (OMAPSPIBusState *) opaque;
     int ch = 0;
     uint32_t ret;
 
+    if (size != 4) {
+        return omap_badwidth_read32(opaque, addr);
+    }
+
     switch (addr) {
     case 0x00:	/* MCSPI_REVISION */
         TRACE("REVISION = 0x%08x", s->revision);
@@ -401,12 +407,16 @@ static uint32_t omap_mcspi_read(void *opaque, target_phys_addr_t addr)
 }
 
 static void omap_mcspi_write(void *opaque, target_phys_addr_t addr,
-                             uint32_t value)
+                             uint64_t value, unsigned size)
 {
     OMAPSPIBusState *s = (OMAPSPIBusState *) opaque;
     uint32_t old;
     int ch = 0;
 
+    if (size != 4) {
+        return omap_badwidth_write32(opaque, addr, value);
+    }
+
     switch (addr) {
     case 0x00:	/* MCSPI_REVISION */
     case 0x14:	/* MCSPI_SYSSTATUS */
@@ -490,7 +500,7 @@ static void omap_mcspi_write(void *opaque, target_phys_addr_t addr,
                 TRACE("invalid TRM value (3)");
             }
                 if (((value >> 7) & 0x1f) < 3) {  /* WL */
-                TRACE("invalid WL value (%i)", (value >> 7) & 0x1f);
+                TRACE("invalid WL value (%" PRIx64 ")", (value >> 7) & 0x1f);
                 }
             if (IS_OMAP3_SPI(s) && ((value >> 23) & 1)) { /* SBE */
                 TRACE("start-bit mode is not supported");
@@ -558,25 +568,19 @@ static void omap_mcspi_write(void *opaque, target_phys_addr_t addr,
                 omap_mcspi_fifo_reset(s);
             }
         } else
-            OMAP_BAD_REGV(addr, value);
+            OMAP_BAD_REG(addr);
         break;
 
     default:
-        OMAP_BAD_REGV(addr, value);
+        OMAP_BAD_REG(addr);
         return;
     }
 }
 
-static CPUReadMemoryFunc * const omap_mcspi_readfn[] = {
-    omap_badwidth_read32,
-    omap_badwidth_read32,
-    omap_mcspi_read,
-};
-
-static CPUWriteMemoryFunc * const omap_mcspi_writefn[] = {
-    omap_badwidth_write32,
-    omap_badwidth_write32,
-    omap_mcspi_write,
+static const MemoryRegionOps omap_mcspi_ops = {
+    .read = omap_mcspi_read,
+    .write = omap_mcspi_write,
+    .endianness = DEVICE_NATIVE_ENDIAN,
 };
 
 static void omap_mcspi_reset(DeviceState *qdev)
@@ -595,7 +599,7 @@ static int omap_mcspi_init(SysBusDevice *busdev)
     OMAPSPIState *s = FROM_SYSBUS(OMAPSPIState, busdev);
     
     s->buscount = (s->mpu_model < omap3430) ? 2 : 4;
-    s->bus = qemu_mallocz(s->buscount * sizeof(*s->bus));
+    s->bus = g_new0(OMAPSPIBusState, s->buscount);
     for (i = 0; i < s->buscount; i++) {
         bs = &s->bus[i];
         if (s->mpu_model < omap3430) {
@@ -607,15 +611,14 @@ static int omap_mcspi_init(SysBusDevice *busdev)
         }
         sysbus_init_irq(busdev, &bs->irq);
         bs->bus = spi_init_bus(&busdev->qdev, NULL, bs->chnum);
-        bs->ch = qemu_mallocz(bs->chnum * sizeof(*bs->ch));
+        bs->ch = g_new0(struct omap_mcspi_ch_s, bs->chnum);
         for (j = 0; j < bs->chnum; j++) {
             sysbus_init_irq(busdev, &bs->ch[j].txdrq);
             sysbus_init_irq(busdev, &bs->ch[j].rxdrq);
         }
-        sysbus_init_mmio(busdev, 0x1000,
-                         cpu_register_io_memory(omap_mcspi_readfn,
-                                                omap_mcspi_writefn, bs,
-                                                DEVICE_NATIVE_ENDIAN));
+        memory_region_init_io(&bs->iomem, &omap_mcspi_ops, bs, "omap.mcspi",
+                              0x1000);
+        sysbus_init_mmio(busdev, &bs->iomem);
     }
     return 0;
 }
@@ -629,20 +632,30 @@ SPIBus *omap_mcspi_bus(DeviceState *qdev, int bus_number)
     hw_error("%s: invalid bus number %d\n", __FUNCTION__, bus_number);
 }
 
-static SysBusDeviceInfo omap_mcspi_info = {
-    .init = omap_mcspi_init,
-    .qdev.name = "omap_mcspi",
-    .qdev.size = sizeof(OMAPSPIState),
-    .qdev.reset = omap_mcspi_reset,
-    .qdev.props = (Property[]) {
-        DEFINE_PROP_INT32("mpu_model", OMAPSPIState, mpu_model, 0),
-        DEFINE_PROP_END_OF_LIST()
-    }
+static Property omap_mcspi_properties[] = {
+    DEFINE_PROP_INT32("mpu_model", OMAPSPIState, mpu_model, 0),
+    DEFINE_PROP_END_OF_LIST()
 };
 
-static void omap_mcspi_register_device(void)
+static void omap_mcspi_class_init(ObjectClass *klass, void *data)
 {
-    sysbus_register_withprop(&omap_mcspi_info);
+    DeviceClass *dc = DEVICE_CLASS(klass);
+    SysBusDeviceClass *k = SYS_BUS_DEVICE_CLASS(klass);
+    k->init = omap_mcspi_init;
+    dc->props = omap_mcspi_properties;
+    dc->reset = omap_mcspi_reset;
 }
 
-device_init(omap_mcspi_register_device)
+static TypeInfo omap_mcspi_info = {
+    .name = "omap_mcspi",
+    .parent = TYPE_SYS_BUS_DEVICE,
+    .instance_size = sizeof(OMAPSPIState),
+    .class_init = omap_mcspi_class_init,
+};
+
+static void omap_mcspi_register_types(void)
+{
+    type_register_static(&omap_mcspi_info);
+}
+
+type_init(omap_mcspi_register_types)
