@@ -18,15 +18,16 @@
  * with this program; if not, see <http://www.gnu.org/licenses/>.
  */
 #include "hw.h"
+#include "qemu-timer.h"
 #include "omap.h"
-
-/* 32-kHz Sync Timer of the OMAP2 */
 struct omap_synctimer_s {
+    MemoryRegion iomem;
     uint32_t val;
     uint16_t readh;
     uint32_t sysconfig; /*OMAP3*/
 };
 
+/* 32-kHz Sync Timer of the OMAP2 */
 static uint32_t omap_synctimer_read(struct omap_synctimer_s *s) {
     return muldiv64(qemu_get_clock_ns(vm_clock), 0x8000, get_ticks_per_sec());
 }
@@ -55,8 +56,8 @@ static uint32_t omap_synctimer_readw(void *opaque, target_phys_addr_t addr)
 static uint32_t omap3_synctimer_readw(void *opaque, target_phys_addr_t addr)
 {
     struct omap_synctimer_s *s = (struct omap_synctimer_s *)opaque;
-    return (addr == 0x04) 
-        ? s->sysconfig 
+    return (addr == 0x04)
+        ? s->sysconfig
         : omap_synctimer_readw(opaque, addr);
 }
 
@@ -67,10 +68,11 @@ static uint32_t omap_synctimer_readh(void *opaque, target_phys_addr_t addr)
 
     if (addr & 2)
         return s->readh;
-
-    ret = omap_synctimer_readw(opaque, addr);
-    s->readh = ret >> 16;
-    return ret & 0xffff;
+    else {
+        ret = omap_synctimer_readw(opaque, addr);
+        s->readh = ret >> 16;
+        return ret & 0xffff;
+    }
 }
 
 static uint32_t omap3_synctimer_readh(void *opaque, target_phys_addr_t addr)
@@ -78,69 +80,79 @@ static uint32_t omap3_synctimer_readh(void *opaque, target_phys_addr_t addr)
     struct omap_synctimer_s *s = (struct omap_synctimer_s *) opaque;
     uint32_t ret;
 
-    if (addr & 2)
+    if (addr & 2) {
         return s->readh;
+    }
 
     ret = omap3_synctimer_readw(opaque, addr);
     s->readh = ret >> 16;
     return ret & 0xffff;
 }
 
-static CPUReadMemoryFunc * const omap_synctimer_readfn[] = {
-    omap_badwidth_read32,
-    omap_synctimer_readh,
-    omap_synctimer_readw,
-};
-
-static CPUReadMemoryFunc * const omap3_synctimer_readfn[] = {
-    omap_badwidth_read32,
-    omap3_synctimer_readh,
-    omap3_synctimer_readw,
-};
-
 static void omap_synctimer_write(void *opaque, target_phys_addr_t addr,
-                                 uint32_t value)
+                uint32_t value)
 {
     OMAP_BAD_REG(addr);
 }
 
+
 static void omap3_synctimer_write(void *opaque, target_phys_addr_t addr,
                                   uint32_t value)
 {
     struct omap_synctimer_s *s = (struct omap_synctimer_s *)opaque;
-    if (addr == 0x04) /* SYSCONFIG */
+    if (addr == 0x04) {
         s->sysconfig = value & 0x0c;
-    else
+    } else {
         OMAP_BAD_REG(addr);
+    }
 }
 
-static CPUWriteMemoryFunc * const omap_synctimer_writefn[] = {
-    omap_badwidth_write32,
-    omap_synctimer_write,
-    omap_synctimer_write,
+static const MemoryRegionOps omap_synctimer_ops = {
+    .old_mmio = {
+        .read = {
+            omap_badwidth_read32,
+            omap_synctimer_readh,
+            omap_synctimer_readw,
+        },
+        .write = {
+            omap_badwidth_write32,
+            omap_synctimer_write,
+            omap_synctimer_write,
+        },
+    },
+    .endianness = DEVICE_NATIVE_ENDIAN,
 };
 
-static CPUWriteMemoryFunc * const omap3_synctimer_writefn[] = {
-    omap_badwidth_write32,
-    omap3_synctimer_write,
-    omap3_synctimer_write,
+static const MemoryRegionOps omap3_synctimer_ops = {
+    .old_mmio = {
+        .read = {
+            omap_badwidth_read32,
+            omap3_synctimer_readh,
+            omap3_synctimer_readw,
+        },
+        .write = {
+            omap_badwidth_write32,
+            omap3_synctimer_write,
+            omap3_synctimer_write,
+        },
+    },
+    .endianness = DEVICE_NATIVE_ENDIAN,
 };
 
 struct omap_synctimer_s *omap_synctimer_init(struct omap_target_agent_s *ta,
-                                             struct omap_mpu_state_s *mpu,
-                                             omap_clk fclk, omap_clk iclk)
+                struct omap_mpu_state_s *mpu, omap_clk fclk, omap_clk iclk)
 {
-    struct omap_synctimer_s *s = qemu_mallocz(sizeof(*s));
+    struct omap_synctimer_s *s = g_malloc0(sizeof(*s));
 
     omap_synctimer_reset(s);
     if (cpu_class_omap3(mpu)) {
-        omap_l4_attach(ta, 0, l4_register_io_memory(omap3_synctimer_readfn,
-                                                    omap3_synctimer_writefn,
-                                                    s));
+        memory_region_init_io(&s->iomem, &omap3_synctimer_ops, s,
+                              "omap.synctimer", omap_l4_region_size(ta, 0));
     } else {
-        omap_l4_attach(ta, 0, l4_register_io_memory(omap_synctimer_readfn,
-                                                    omap_synctimer_writefn,
-                                                    s));
+        memory_region_init_io(&s->iomem, &omap_synctimer_ops, s,
+                              "omap.synctimer", omap_l4_region_size(ta, 0));
     }
+    omap_l4_attach(ta, 0, &s->iomem);
+
     return s;
 }
