@@ -8,7 +8,7 @@
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
  * published by the Free Software Foundation; either version 2 or
- * (at your option) any later version of the License.
+ * (at your option) version 3 of the License.
  *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -315,7 +315,7 @@ struct MUSBEndPoint {
 };
 
 struct MUSBState {
-    qemu_irq *irqs;
+    qemu_irq irqs[musb_irq_max];
     USBBus bus;
     USBPort port;
 
@@ -331,9 +331,7 @@ struct MUSBState {
     uint16_t rx_intr;
     uint16_t rx_mask;
 
-#ifdef SETUPLEN_HACK
     int setup_len;
-#endif
     int session;
 
     uint8_t buf[0x8000];
@@ -350,7 +348,6 @@ void musb_reset(MUSBState *s)
     s->faddr = 0x00;
     s->devctl = 0;
     s->power = MGC_M_POWER_HSENAB;
-
     s->tx_intr = 0x0000;
     s->rx_intr = 0x0000;
     s->tx_mask = 0xffff;
@@ -359,9 +356,7 @@ void musb_reset(MUSBState *s)
     s->mask = 0x06;
     s->idx = 0;
 
-#ifdef SETUPLEN_HACK
     s->setup_len = 0;
-#endif
     s->session = 0;
     memset(s->buf, 0, sizeof(s->buf));
 
@@ -380,12 +375,13 @@ void musb_reset(MUSBState *s)
 
 struct MUSBState *musb_init(DeviceState *parent_device, int gpio_base)
 {
-    MUSBState *s = qemu_mallocz(sizeof(*s));
-    s->irqs = qemu_mallocz(__musb_irq_max * sizeof(qemu_irq));
+    MUSBState *s = g_malloc0(sizeof(*s));
     int i;
-    for (i = 0; i < __musb_irq_max; i++) {
+
+    for (i = 0; i < musb_irq_max; i++) {
         s->irqs[i] = qdev_get_gpio_in(parent_device, gpio_base + i);
     }
+
     musb_reset(s);
 
     usb_bus_new(&s->bus, &musb_bus_ops, parent_device);
@@ -610,6 +606,8 @@ static int musb_timeout(int ttype, int speed, int val)
 static void musb_packet(MUSBState *s, MUSBEndPoint *ep,
                 int epnum, int pid, int len, USBCallback cb, int dir)
 {
+    USBDevice *dev;
+    USBEndpoint *uep;
     int ret;
     int idx = epnum && dir;
     int ttype;
@@ -627,16 +625,14 @@ static void musb_packet(MUSBState *s, MUSBEndPoint *ep,
     ep->delayed_cb[dir] = cb;
 
     /* A wild guess on the FADDR semantics... */
-    usb_packet_setup(&ep->packey[dir].p, pid, ep->faddr[idx],
-                     ep->type[idx] & 0xf);
+    dev = usb_find_device(&s->port, ep->faddr[idx]);
+    uep = usb_ep_get(dev, pid, ep->type[idx] & 0xf);
+    usb_packet_setup(&ep->packey[dir].p, pid, uep);
     usb_packet_addbuf(&ep->packey[dir].p, ep->buf[idx], len);
     ep->packey[dir].ep = ep;
     ep->packey[dir].dir = dir;
 
-    if (s->port.dev)
-        ret = usb_handle_packet(s->port.dev, &ep->packey[dir].p);
-    else
-        ret = USB_RET_NODEV;
+    ret = usb_handle_packet(dev, &ep->packey[dir].p);
 
     if (ret == USB_RET_ASYNC) {
         ep->status[dir] = len;
@@ -817,7 +813,8 @@ static void musb_async_cancel_device(MUSBState *s, USBDevice *dev)
 
     for (ep = 0; ep < 16; ep++) {
         for (dir = 0; dir < 2; dir++) {
-            if (s->ep[ep].packey[dir].p.owner != dev) {
+            if (!usb_packet_is_inflight(&s->ep[ep].packey[dir].p) ||
+                s->ep[ep].packey[dir].p.ep->dev != dev) {
                 continue;
             }
             usb_cancel_packet(&s->ep[ep].packey[dir].p);
@@ -1324,7 +1321,7 @@ static void musb_writeb(void *opaque, target_phys_addr_t addr, uint32_t value)
         s->power = (value & 0xef) | (s->power & 0x10);
         /* MGC_M_POWER_RESET is also read-only in Peripheral Mode */
         if ((value & MGC_M_POWER_RESET) && s->port.dev) {
-            usb_send_msg(s->port.dev, USB_MSG_RESET);
+            usb_device_reset(s->port.dev);
             /* Negotiate high-speed operation if MGC_M_POWER_HSENAB is set.  */
             if ((value & MGC_M_POWER_HSENAB) &&
                             s->port.dev->speed == USB_SPEED_HIGH)
@@ -1545,7 +1542,7 @@ static void musb_writew(void *opaque, target_phys_addr_t addr, uint32_t value)
         musb_write_fifo(s->ep + ep, (value >> 8 ) & 0xff);
         musb_write_fifo(s->ep + ep, (value >> 16) & 0xff);
         musb_write_fifo(s->ep + ep, (value >> 24) & 0xff);
-        break;
+            break;
     default:
         TRACE("unknown register 0x%02x", (int) addr);
         break;
